open Rtype
open RSyntax
open Closure

(* $BCj>](B  C $B8@8l%3!<%I(B *)


type ident = string
type label = string
type index = int
type size = int


let new_ident = 
  let n = ref 0 in function () -> n := !n+1; "ac" ^ (string_of_int !n)


type ac_exp = 
    AC_Val of rvalue
  | AC_Var of ident
  | AC_refFv of index
  | AC_refFrv of index
  | AC_ApplyClosure of ident * ident list
  | AC_ApplyRClosure of ident * region list
	
type ac_statement = 
    AC_If of ident * (ac_statement list) * (ac_statement list)
  | AC_Subst of ident * ac_exp
  | AC_fvSubst  of ident(*closure*) * index * ac_exp 
  | AC_rfvSubst of ident(*closure*) * index * region 
  | AC_Return of ac_exp
  | AC_openRegion of region
  | AC_closeRegion of region
  | AC_InitClosure of (ident * (label * size(*fv*) * size(*frv*) * region))
  | AC_InitRClosure of (ident *  (label * size(*fv*) * size(*frv*) * region))

(* $B4X?tDj5A(B *)
type ac_defun = {ac_fname : ident;  
		 ac_args    : arg_format;
		 ac_body : ac_statement list;
		   
	       }
(* $B4X?tDj5A(B *)
type ac_rdefun = {ac_rfname : ident;
		  ac_rargs    : region list;
		  ac_rbody : ac_statement list 
	       }



type ac_main = ac_statement list

type ac_program = ac_defun list * ac_rdefun list * ac_main



type closure_info = {clo_name:ident; 
		     clo_fv:ident list;
		     clo_frv:region list;
		     clo_reg:region}

(* closure $B$N(B body $BItJ,$rL?NaNs$KJQ49$9$k(B *)
let cexp_to_ac_statement_list (clo:closure_info) ce = 
  (* ce$B$NJV$9CM$r(B v $B$KBeF~$9$k(B *)
  let rec iter v ce =
    match ce with
      CVal(v) -> [AC_Subst(new_ident (),AC_Val(v))]
    | CVar(v) -> 
	if List.mem v clo.clo_fv 
	then 
	  [AC_Subst(new_ident (),
		    AC_refFv(Mylist.indexOf v clo.clo_fv))]
	else 
	  [AC_Subst(new_ident (),AC_Var(v))]
    | CLet(b,body) -> 
	(iter b.cid b.cbody)@(iter v body)
    | CMake_Closure(var,(name,fvs,rfvs,r),body) ->
	(AC_InitClosure(var,(name,List.length fvs, List.length rfvs, r))) (* $B=i4|2=(B *)
	:: (let rec iter n lst =
	  match lst with
	    hd::tl -> 
	      (AC_fvSubst(var, n, AC_Var(hd)))::(iter (n+1) tl)
	  | [] -> []
	in iter 0 fvs)
	@ (let rec iter n lst =
	    match lst with
	      hd::tl -> 
		(AC_rfvSubst(var, n, hd))::(iter (n+1) tl)
	    | [] -> []
	  in iter 0 rfvs)
	@ (iter v body)
    | CMake_RClosures(clos,body) ->
	(* $B=i4|2=(B *)
	(List.map 
	  (function 
	      RClosure(var,(l,ids,rs,reg)) -> 
		AC_InitRClosure(var,(l,List.length ids, List.length rs, reg))
	    | RCVal _ -> failwith "error in abstractC.ml") clos)
	@
	  (List.fold_right (@)
	     (List.map 
		(function  RClosure(var,(l,ids,rs,reg)) ->
		  (let rec iter n lst =
		    match lst with
		      hd::tl -> 
			(AC_fvSubst(var, n, AC_Var(hd)))::(iter (n+1) tl)
		    | [] -> []
		  in iter 0 ids)
		  @
		    (let rec iter n lst =
		      match lst with
			hd::tl -> 
			  (AC_rfvSubst(var, n, hd))::(iter (n+1) tl)
		      | [] -> [] 
		    in iter 0 rs) 
		  | RCVal _ -> failwith "error in abstractC" ) clos) [])
	@ (iter v body)
        
    | CApply_Closure(id,args) -> 
	[AC_Subst(v,AC_ApplyClosure(id,args)) ]
    | CRApply_Closure(id,regs) -> 
	[AC_Subst(v,AC_ApplyRClosure(id,regs)) ]
    | CIf(id,e1,e2) ->
	[AC_If(id,iter v e1,iter v e2)]
    | CLetregion(regs,body) ->
	(List.map (function x -> AC_openRegion(x)) regs)
	@ (iter v body)
	@ (List.map (function x -> AC_closeRegion(x)) regs)

  in 
  let v = new_ident () in
  (iter v ce)@[AC_Return(AC_Var v)]


let closure_def2ac_defun cdef = 
  let cloinfo = { clo_name = cdef.label;
		  clo_fv = cdef.fvars;
		  clo_frv = cdef.frvars;
		  clo_reg = cdef.reg } in
  let stmts = cexp_to_ac_statement_list cloinfo cdef.body in
  {ac_fname   = cdef.label;
   ac_args    = cdef.args;
   ac_body    = stmts }


let rclosure_def2ac_rdefun cdef = 
  let cloinfo = { clo_name = cdef.r_label;
		  clo_fv = cdef.r_fvars;
		  clo_frv = cdef.r_frvars;
		  clo_reg = cdef.r_reg } in
  let stmts = cexp_to_ac_statement_list cloinfo cdef.r_body in
  {ac_rfname   = cdef.r_label;
   ac_rargs    = cdef.r_rargs;
   ac_rbody    = stmts }
   

let clo_to_ac_program (clodefs,rclodefs,main) = 
  (List.map closure_def2ac_defun clodefs,
   List.map rclosure_def2ac_rdefun rclodefs,
   let cloinfo = { clo_name = "__dummy__";
		   clo_fv = [];
		   clo_frv = [];
		   clo_reg = Rtype.newRegVar ()} in
   cexp_to_ac_statement_list cloinfo main)
  
(*  closure_def_to_ac_defun (clodefs *)
  


(* ========================================================= *)

open Format

let rec create_arglst args = 
  match args with
    [hd] -> hd
  | hd::tl-> hd ^ "," ^ (create_arglst tl)
  | [] -> ""

let rec vars_to_argdef arg = 
  let rec iter lst =
    match lst with
      [hd] -> "ScmVal* " ^ hd 
    | hd::tl -> ("ScmVal* " ^ hd ^ "," ^ iter tl)
    | [] -> ""
  in 
  match arg with
    (ids,None) ->
      iter ids
  | (ids,_ ) -> failwith "not support"

let rec vars_to_clodef arg = 
  match arg with
    [hd] -> "Region* " ^ hd 
  | hd::tl -> ("Region* " ^ hd ^ "," ^ vars_to_clodef tl)
  | [] -> ""
	  
let rec c_code_of_rvalue rval = 
  match rval with
    RNumber(Library.S_Int(i),reg) -> 
      "new_ScmInt(int" ^ (string_of_int i) 
      ^ "," ^ (string_of_reg reg) ^ ")"
  | _ -> string_of_rvalue rval
	
	
let print_abstractC (defunlst, rdefunlst, main) =
  let rec pr_ac_exp ppf acexp = 
    match acexp with
      AC_Val(rv) -> fprintf ppf "%s" (c_code_of_rvalue rv)
    | AC_Var(id) -> fprintf ppf "%s" id
    | AC_refFv(idx) -> fprintf ppf "get_fvar(closure,%d)" idx
    | AC_refFrv(idx) -> fprintf ppf "get_frvar(closure,%d)" idx
    | AC_ApplyClosure(id,args) -> 
	fprintf ppf "%s.f(%s)" id (create_arglst args)
    | AC_ApplyRClosure(id,regs) ->
	fprintf ppf "%s.f(%s)" id (create_arglst (List.map string_of_reg regs))
  and  pr_ac_stmt ppf stmt = 
    match stmt with
      AC_If(id,stlst1,stlst2) ->
	fprintf ppf "@[<hov 0>if(ScmTest(%s)){@, %a@,}else{@, %a@,}@,@]" 
	  id pr_ac_stmtlst stlst1 pr_ac_stmtlst stlst2
    | AC_Subst(id,exp)  ->
	fprintf ppf "@[<hov 0>ScmVal* %s = %a;@]" id pr_ac_exp exp
    | AC_fvSubst(id,idx,exp) ->
	fprintf ppf "@[<hov 0>set_fvar(%s,%d,%a);@]" id idx  pr_ac_exp exp
    | AC_rfvSubst(id,idx,reg) ->
	fprintf ppf "@[<hov 0>set_frvar(%s,%d,%s);@]" id idx  (string_of_reg reg)
    | AC_Return(exp) ->
	fprintf ppf "@[<hov 0>return %a;@]" pr_ac_exp exp
    | AC_openRegion(reg) ->
	fprintf ppf "@[<hov 0>Region* %s = new_region();@]" (string_of_reg reg)
    | AC_closeRegion(reg) ->
	fprintf ppf "@[<hov 0>delete_region(%s);@]" (string_of_reg reg)
    | AC_InitClosure(var,(label,size1,size2,reg)) ->
	fprintf ppf "@[<hov 0>ScmVal* %s =  new_ScmProc_alloc(%s,%d,%d,%s);@]" 
	  var label size1 size2 (string_of_reg reg)
    | AC_InitRClosure(var,(label,size1,size2,reg)) ->
	fprintf ppf "@[<hov 0>ScmVal* %s =  new_ScmProc_alloc(%s,%d,%d,%s);@]" 
	  var label size1 size2 (string_of_reg reg)
  and pr_ac_stmtlst ppf stmtlst = 
    match stmtlst with
      hd::tl ->
	fprintf ppf "@[<hov 0>%a@\n%a@]" pr_ac_stmt hd pr_ac_stmtlst tl
    | [] -> 
	fprintf ppf "" 
  and pr_ac_defun ppf defun =
    let v_args = 
      match defun.ac_args with
	([],None) -> "ScmVal* closure, int length"
      | _ -> "ScmVal* closure, int length," 
	  ^ (vars_to_argdef defun.ac_args) in
    fprintf ppf "@[<v 0>ScmVal* %s(%s){@,  %a@,}@]" 
      defun.ac_fname v_args pr_ac_stmtlst defun.ac_body

  and pr_ac_defun_lst ppf defunlst =
    match defunlst with
      hd::tl ->
	fprintf ppf "@[<hv 0>%a@\n%a@]" pr_ac_defun hd pr_ac_defun_lst tl
    | [] -> 
	fprintf ppf "" 
    
    
  and pr_ac_rdefun ppf defun =
    let v_args = 
      match defun.ac_rargs with
	[] -> "ScmVal* closure, int length"
      | _ -> "ScmVal* closure,int length,"
	  ^ (vars_to_clodef (List.map string_of_reg defun.ac_rargs)) in
    fprintf ppf "@[<v 0>ScmVal* %s(%s){@,  %a@,}@]" 
      defun.ac_rfname v_args pr_ac_stmtlst defun.ac_rbody
      
  and pr_ac_rdefun_lst ppf defunlst =
    match defunlst with
      hd::tl ->
	fprintf ppf "@[<hov 0>%a@\n%a@]" pr_ac_rdefun hd pr_ac_rdefun_lst tl
    | [] -> 
	fprintf ppf "" 
    
  and pr_main ppf main =
    fprintf ppf "@[<hov 0>int main(){@\n %a @\n}@\n@]" pr_ac_stmtlst main

  in 
  printf "@[<hov 0>%a@\n%a@\n%a@\n@]" 
    pr_ac_rdefun_lst rdefunlst pr_ac_defun_lst defunlst   pr_main main
    
	  
