(* 르ꥺRμ *)

open RSyntax
open Rtype

let debug str = () (* (print_endline ("algR:"^str); flush_all ())*)

let size = ref 0
let progress = 
  let c = ref (-1) in 
  (function () -> c := !c + 1; print_endline ((string_of_int !c) ^ "/" ^ ( string_of_int !size)); flush_all ())

(** 르ꥺRο *)
let rec inferR (renv:rtype_env) (re:rexp) : (rsubst * rexp) =
  (* progress (); *)
(*  print_string "naze --";
  simple_print_rexp re;*)
  match re with
    RVal(_,_,_) -> debug "val";(rsubst_id, re)
  | RVar(_,_,_) -> debug "var";(rsubst_id, re)
  | RFVar(v, rlst, elst, rt, ef) -> 
      debug "rfvar";
      begin
	match (List.assoc v renv) with
	  Normal(_) -> failwith "AlgorithmR : never occur"
	| Forall(reglst,peff, efflst, srt, reg) ->
(*	    if(v = "v11")
	    then 
	      begin
		print_rexp re;
		print_endline (string_of_rtype_scheme (List.assoc v renv));
		(Printf.printf "aaaaa %s : %d(hiki) = %d(type) , %d = %d\n" 
		   v
		   (List.length rlst) (List.length reglst)
		   (List.length elst) (List.length efflst))
	      end 
	    else 
	      (); *)
	    let newS = (list_map_to_rsubst (List.combine reglst rlst),
			List.combine efflst 
			  (List.map (function e -> (e,[])) elst)) in
	    let newrt = substToType newS srt in
	    let rs = unifyRtype newrt rt in
	    (* print_endline ("---" ^ v) ;
	    print_endline (string_of_rtype rt);
	    print_endline (string_of_rtype newrt); *)
	    let rt2 = substToType rs rt in
	    (* print_endline (string_of_rtype rt2);
	    print_endline (string_of_rtype (substToType rs rt2)); *)
	    let ret_eff = (Reg reg)::(substToEffect 
					(rsubstCombine rs newS)
					(regpolEffToEff peff)) in
	    (rs, substToRexp rs (RFVar(v,rlst,elst,
				       rt2, sortEffect ret_eff)))
      end
  | RLambda((args,tail), body, reg, rt, eff) ->
      debug "lambda";
      let newEnv = 
	begin
	  match rt.func with
	    None -> failwith "AlgorithmR.inferR : Invalid Argument at RLambda ."
	  | Some((t_arg,t_tail), t_eff, t_ret, t_reg) ->
	      let rec iter args t_args =
		begin 
		  match (args,t_args) with
		    (ahd::atl,thd::ttl) ->
		      (ahd,Normal thd)::(iter atl ttl)
		  | (ahd::atl,[]) -> (ahd,Normal(rtype_car_of_pair t_tail))::(iter atl [])
		  | ([],thd::ttl) -> 
		      failwith "AlgorithmR.inferR : Invalid Argument at RLambda ."
		  | ([],[]) -> 
		      begin 
			match tail with
			  Some t -> [t,Normal t_tail]
			| None -> []
		      end
		end 
	      in 
	      iter args t_arg
	end
      in 
      let (rs,new_body) = inferR (newEnv@renv) body in
      (rs,
       RLambda((args,tail), new_body, 
	       substToRegion rs reg,
	       substToType rs rt,
	       sortEffect (substToEffect rs eff)))
  | RApp(rf, args, rt, eff) ->
      debug "app";
      let (rs, new_rf) = inferR renv rf in
      let rec inferArgs s args =
	begin 
	  match args with
	    hd::tl -> 
	      let (tmpS,rhd) = inferR (substToRenv s renv) (substToRexp s hd)  in
	      let (tails,rtail) = inferArgs (rsubstCombine tmpS s) tl in
	      (tails,substToRexp tails rhd::rtail)
	  | [] -> (s,[])
	end
      in 
      let (r_subst,rargs) = inferArgs rs args in
      (r_subst,
       RApp(substToRexp r_subst rf,
	    rargs (* List.map (substToRexp r_subst) args*),
	    substToType r_subst rt,
	    sortEffect (substToEffect r_subst eff)))
  | RLet(rbinds, body, rt ,eff) ->
      debug "let";
      let rec inferRbinds s rbinds =
	begin 
	  match rbinds with
	    hd::tl -> 
	      let (tmpS,_) = 
		inferR (substToRenv s renv) (substToRexp s hd.rbody) in
	      inferRbinds tmpS tl
	  | [] -> s
	end
      in 
      let bindS = inferRbinds rsubst_id rbinds in
      let newenv = List.map (function x -> (x.rid,x.rt)) rbinds in
      let (rS,rbody) = inferR (substToRenv bindS (newenv@renv)) 
	  (substToRexp bindS body) in
      (rS,
       RLet(List.map (function x -> {rid=x.rid; 
				     rt=substToRScheme rS x.rt;
				     rbody=substToRexp rS x.rbody})
	      rbinds,
	    rbody,
	    substToType rS rt,
	    sortEffect (substToEffect rS eff)))
  | RIf(re1, re2, re3, rt, eff) ->
      debug "if";
      let (tmpS,tr1) = inferR renv re1 in      
      let (tmpS2,tr2) = inferR (substToRenv tmpS renv) (substToRexp tmpS re2) in
      let tmpS2' = rsubstCombine tmpS2 tmpS in
      let (tmpS3,tr3) = inferR (substToRenv tmpS2' renv) (substToRexp tmpS2' re3) in
      let tmpS3' = rsubstCombine tmpS3 tmpS2' in
      (tmpS3',
       RIf(substToRexp tmpS3' tr1,
	   substToRexp tmpS3' tr2,
	   substToRexp tmpS3' tr3,
	   substToType tmpS3' rt,
	   sortEffect (substToEffect tmpS3' eff)))
  | RLetrec(rbinds, rbody, rt, eff) ->
      debug "letrec";
      (* inferR κƵ *)
      let rec inferRrec (rs:rsubst) (crnt_binds:rbind list)  =
	let crnt_env = List.map (function x -> (x.rid, x.rt)) crnt_binds in

	let rec inferBinds (rs:rsubst) rbinds =
	  match rbinds with
	    bnd::tl -> 
	      let (tmpS, new_rbody) = 
		inferR (substToRenv rs (crnt_env@renv)) (substToRexp rs bnd.rbody) in
	      let (tlS,new_tl) = inferBinds (rsubstCombine tmpS rs) tl in
	      (tlS, (bnd.rid, substToRexp tlS new_rbody)::new_tl)
	  | [] -> (rs,[])
	in 
	let (tmpS, tmpbinds) = inferBinds rs crnt_binds in
	let newbinds = 
	  List.map 
	    (function ((_,rs),(id,rt)) -> 
	      {rid=id; 
	       rt=
	       if(isRValue rt) 
	       then
		 begin 
		   let rt = rgeneralize 
		       (substToRenv tmpS (crnt_env@renv)) 
		       (getRtype (substToRexp tmpS rt))
		       begin match rs with
			 Forall(_,_,_,_,polreg) -> (substToRegion tmpS polreg )
		       | _ -> failwith "AlgorithmR.inferR : error "
		       end
		   in 
		   rt
		 end
	       else 
		 Normal(getRtype (substToRexp tmpS rt))
		   ;
	       rbody= substToRexp tmpS rt})
	    (List.combine crnt_env tmpbinds) in
	(* ब fixed-pointޤãΤ롣*)
	if (List.fold_right (&&)
	      (List.map 
		 (function (old_bnd, new_bnd) -> 
		   isEquivRtypeScm old_bnd.rt new_bnd.rt) 
		 (List.combine crnt_binds newbinds)) true)
	then (* Ƶλ *)
	  (tmpS, newbinds) 
	else (* ⤦ *)
	  begin 
	    let replace_all body = 
	      List.fold_right 
		(fun (old_b,new_b) b -> 
		  replaceRscheme new_b.rid old_b.rt new_b.rt b)
		(List.combine crnt_binds newbinds) body
	    in 
	    let rep_binds = 
	      (List.map 
		 (function (old_bnd, new_bnd) -> 
		   {rid=new_bnd.rid;
		    rt=new_bnd.rt;
		    rbody= replace_all new_bnd.rbody
		  } )
		 (List.combine crnt_binds newbinds))
	    in 
	    inferRrec (rsubstCombine tmpS rs) rep_binds
	  end
	    (* Ƶޤ *)
      in 
      let (bndS, newBinds) = inferRrec rsubst_id rbinds in
      let bnd_env = List.map (function x -> (x.rid, x.rt)) newBinds in
      (* Τ꡼ΰθĿѹ *)
      let rec reset_regargs old_rbinds new_rbinds rbody = 
	match (old_rbinds,new_rbinds) with
	  (ohd::otl,nhd::ntl) ->
	    replaceRscheme nhd.rid ohd.rt nhd.rt (reset_regargs otl ntl rbody)
	| ([],[]) -> rbody
	| _ -> failwith "invalid argument :  reset rbody"
      in 
      let tmpRbody = reset_regargs (List.map (substToRbind bndS) rbinds) newBinds (substToRexp bndS rbody) in
      let (lrS, new_rbody) = inferR (substToRenv bndS (bnd_env@renv)) tmpRbody in
      let lastS = rsubstCombine lrS bndS in
      let lastBinds = 
	List.map (function x -> {rid=x.rid;
				 rt=substToRScheme lastS x.rt;
				 rbody=substToRexp lastS x.rbody}) newBinds in
      let getEffect_of_rbind (rbnd:rbind) : effect = 
	match rbnd.rt with
	  Forall(_,_,_,_,r) -> [Reg(r) ]
	| Normal(rt) -> getEffect rbnd.rbody
      in 

      (lastS,
       RLetrec(lastBinds,
	       new_rbody, 
	       getRtype new_rbody,
	       sortEffect (List.fold_right (@) 
			     (List.map getEffect_of_rbind lastBinds)
			     (getEffect new_rbody))))
	
  | RLetregion(regs, re, rt, eff) ->
      debug "letregion";
      let (tmpS, new_re) = inferR renv re in
      (tmpS,insertLetRegion (substToRenv tmpS renv) new_re)
	
