open TSyntax
open Rtype
open RSyntax

let debug str = () (*(print_endline ("algS:"^str); flush_all ())*)

let size = ref 0
let progress = 
  let c = ref (-1) in 
  (function () -> ())
(*    c := !c + 1; 
    print_endline ((string_of_int !c) ^ "/" ^ ( string_of_int !size)); 
    flush_all ()) *)


let print_need_vars (lreg,leff) =
  List.iter print_string (List.map string_of_reg lreg);
  print_string " + ";
  List.iter print_string leff;
  print_endline ""

(* 르ꥺ *)
let rec inferS ((livreg,liveff):(region list * effvar list)) (renv:rtype_env) (t:texp) : (rsubst * rexp) =

  progress ();
  match t with
    TVal(Syntax.Number(i),st) ->  debug "val num";
      let rt1 = addRegEff st in
      begin 
	match rt1.num with Some(r1) ->
	  (rsubst_id, RVal(RNumber(i,r1),rt1,[Reg(r1)]))
	| None -> failwith "RSintax.infreS : Unknown exception (num)"
      end
  | TVal(Syntax.Boolean(b),st) -> debug "val bool";
      let rt1 = addRegEff st in
      begin match rt1.bool with true ->
	    (rsubst_id, RVal(RBoolean(b),rt1,[]))
	| false ->
	    failwith "RSintax.infreS : Unknown exception (bool)"
      end
  | TVal(Syntax.Null,st) -> debug "val null";
      let rt1 = addRegEff st in
      begin match rt1.null with true ->
	    (rsubst_id, RVal(RNull,rt1,[]))
	| false ->
	    failwith "RSintax.infreS : Unknown exception (null)"
      end
  | TVal(Syntax.Undef,st) -> debug "val undef";
      let rt1 = addRegEff st in
      (rsubst_id, RVal(RUndef,rt1,[]))
  | TVal(Syntax.Symbol(sym),st) -> debug "val symbol";
      let rt1 = addRegEff st in
      begin 
	match rt1.symbol with Some(r1) ->
	  (rsubst_id, RVal(RSymbol(sym,r1),rt1,[Reg(r1)]))
	| None -> failwith "RSintax.infreS : Unknown exception (symbol)"
      end
  | TVal(Syntax.String(sym),st) -> debug "val symbol";
      let rt1 = addRegEff st in
      begin 
	match rt1.string with Some(r1) ->
	  (rsubst_id, RVal(RString(sym,r1),rt1,[Reg(r1)]))
	| None -> failwith "RSintax.infreS : Unknown exception (string)"
      end
  | TVal(Syntax.Character(c),st) -> debug "val symbol";
      let rt1 = addRegEff st in
      begin 
	match rt1.char with Some(r1) ->
	  (rsubst_id, RVal(RCharacter(c,r1),rt1,[Reg(r1)]))
	| None -> failwith "RSintax.infreS : Unknown exception (string)"
      end
(*  | TVal(_,_) -> failwith "RSintax.infreS : Yet not implemented value"*)
  | TVar(x,st) -> debug ("var" ^x);
      begin
	match (List.assoc x renv) with
	  Forall(regs,eff, effs, rt, r) ->
	    let newregs = List.map (function a -> newRegVar ()) regs in
	    let neweffs = List.map (function a -> newLatentEffect ()) effs in
	    let tmpS = (list_map_to_rsubst (List.combine regs newregs), 
			List.combine effs neweffs) in
	    let newrt = instantiate tmpS rt st in
	    let ret_effect =  (Reg r)::(substToEffect tmpS (regpolEffToEff eff)) in
	    (rsubst_id,RFVar(x, newregs, List.map fst neweffs,
			     newrt, ret_effect))
	| Normal(rtype) ->
	    (* ̾ηѿ *)
	    debug (string_of_rtype rtype);
	    (rsubst_id,RVar(x,rtype,[]))
      end
  | TLambda((lblst,tail),te,st) -> debug "lambda";
      (*  ꡼ղä롣*)
      let newt = addRegEff st in
	begin
	  match newt.func with
	    None -> 
	      failwith "RSyntax.inferS : Invalid argument at TLamda. "
	  | Some((args,tailT),ef,ret,reg) ->
	      (* Ķ *)
	      let rec getNewEnv lblst args = 
		begin
		  match (lblst,args) with
		    (lhd::ltl,ahd::atl) ->
		      (lhd.lid,Normal ahd)::(getNewEnv ltl atl)
		  | (lhd::ltl,[]) -> 
		      (lhd.lid,Normal (rtype_car_of_pair tailT))::(getNewEnv ltl [])
		  | ([],_) -> []
		end
	      in 
	      let argEnv = getNewEnv lblst args in
	      let newEnv = 
		begin
		  match tail with 
		    Some t -> 
		      (t.lid,Normal(rtype_car_of_pair tailT))::argEnv 
		  | None -> argEnv
		end
	      in
	      (* ret  *)
	      let (lreg,leff) = fv_of_Env newEnv in
	      let (tmpS,rret) = inferS (lreg@livreg,leff@liveff) (newEnv@renv) te in
	      let (tmpS2) = unifyRtype (getRtype rret) (substToType tmpS ret) in
	      let tmpS2' = rsubstCombine tmpS2 tmpS in
	      (* եȤ碌 *)
	      let tmpS3 = unifyLatentEffect (substToLatentEffect tmpS2' ef)
		  (newEffVar (), sortEffect (getEffect (substToRexp tmpS2' rret))) in
	      let tmpS3' = rsubstCombine tmpS3 tmpS2' in
	      let rtail = 
		begin match tail with Some t -> Some t.lid | None -> None end in
	      (tmpS3',
	       RLambda((List.map fst argEnv,
			rtail),
		       substToRexp tmpS3' rret,
		       substToRegion tmpS3' reg,
		       substToType tmpS3' newt,
		       substToEffect tmpS3' [Reg(reg)]))
	end
  | TApp(f,args,stype) -> debug "app";
      (* ؿΤ򷿿 *)
      let (sf,rf) = inferS (livreg,liveff) renv f in
      (* 򷿿 *)
      let rec inferArgs (livreg,liveff) (crntS:rsubst) (args:texp list) = 
	begin
	  match args with
	    hd::tl -> 
	      let (hdS, rhd) = inferS (livreg,liveff) (substToRenv crntS renv) hd in
	      let (lreg,leff) = fv_of_rexp rhd in
	      let (tlS, rtl) = inferArgs (lreg@livreg,leff@liveff) (rsubstCombine hdS crntS) tl in
	      (tlS, rhd::rtl) 
	  | [] -> (crntS, [])
	end
      in 
      let (argS, rargs) = 
	inferArgs (Mylist.append2 (fv_of_rexp rf) (livreg,liveff)) sf args in
      
      begin 
	match (getRtype rf).func with
	  None -> failwith "AlgorithmS.inferS : Invalid argument at TApp. "
	| Some((t_args,t_tail), ef, t_ret, reg) ->
	    (*  unify *)
	    let rec unifyArgs rargs t_args (s:rsubst) =
	      match (rargs,t_args) with
		(rhd::rtl, thd::ttl) ->
		  let new_s = unifyRtype 
		      (substToType s thd) 
		      (substToType s (getRtype rhd))  in
		  unifyArgs rtl ttl (rsubstCombine new_s s)
	      | (rhd::rtl, []) -> 
		  let new_s = unifyRtype 
		      (substToType s (rtype_car_of_pair t_tail))
		      (substToType s (getRtype rhd))  in
		  unifyArgs rtl [] (rsubstCombine new_s s)
	      | ([],thd::ttl) -> failwith "AlgorithmS.inferS : Invalid argument at TApp. "
	      | ([],[]) -> s
	    in 
	    let argS2 = unifyArgs rargs t_args argS in
	    (* print_endline (string_of_rsubst argS2); *)
	    let fe = 
	      insertLetRegion 
		(substToRenv argS2 renv) 
		(RApp(substToRexp argS2 rf,
		      List.map (substToRexp argS2) rargs,
		      substToType argS2 t_ret,
		      sortEffect (substToEffect argS2
				    ((flatten_Leff ef)
				     @ (getEffect rf)
				     @ (List.fold_right (@) (List.map getEffect rargs) []))
				 )))
	    in 
	    let appS = subst_filter (livreg,liveff) argS2 in
	    (argS2,fe)
      end
	
  | TLet(tblst,te,st) -> debug "let";
      let inferForTbind lives (s:rsubst) (tb:tbind) : (rsubst * rbind) =
	let (newS,rb) = inferS lives (substToRenv s renv) tb.sbody in
	if isRValue(rb) 
	then 
	  let rscm = rgeneralize (substToRenv s renv) (getRtype rb) (newRegVar ()) in
	  (newS,  {rid = tb.sid; rt=rscm; rbody=rb })
	else 
	  (newS,  {rid = tb.sid; rt=Normal(getRtype rb); rbody=rb })
      in 
      (* «ʬο *)
      let rec iter (livreg,liveff) (tblst:tbind list) (s:rsubst) : (rsubst * rbind list) =
	match tblst with
	  hd::tl ->
	    let (newS,rb) = inferForTbind (livreg,liveff) s hd in
	    let (lreg,leff) = fv_of_rexp rb.rbody in
	    let (newS2,rtl) = iter (lreg@livreg,leff@liveff) tl (rsubstCombine newS s) in
	    (newS2, 
	     {rid = rb.rid; rt=substToRScheme newS2 rb.rt; 
	      rbody=substToRexp newS2 rb.rbody} :: rtl)
	| [] -> (s,[])
      in 
      let (bindS,rbinds) = iter (livreg,liveff) tblst rsubst_id in
      let newEnv = List.map (function x -> (x.rid , x.rt)) rbinds in
      let (lreg,leff) = List.fold_right Mylist.append2 (List.map (function b -> fv_of_rexp b.rbody) rbinds) ([],[]) in
      let (es,re) = inferS (lreg@livreg, leff@liveff) (newEnv@(substToRenv bindS renv)) te in
      let rbinds' = 
	List.map (function x -> {rid=x.rid; rt=substToRScheme es x.rt;
				 rbody=substToRexp es x.rbody}) rbinds
      in 
      let getEffect_of_rbind (rbnd:rbind) : effect = 
	match rbnd.rt with
	  Forall(_,_,_,_,r) -> [Reg(r) ]
	| Normal(rt) -> getEffect rbnd.rbody
      in 
      let letS = rsubstCombine es bindS in
      (letS,
       insertLetRegion 
	 (substToRenv letS renv)
	 (RLet(rbinds',
	       re,
	       getRtype re,
	       sortEffect (List.fold_right (@) 
			     (List.map getEffect_of_rbind rbinds')
			     (getEffect re)))))
  | TLetrec(tblst,te,st) ->debug "letrec";
      (* Ķ  *)
      let makeNewEnvTbind (tb:tbind) : (string * rtype_scheme) =
	if(TSyntax.isValue tb.sbody)
	then 
	  (tb.sid, 
	   rgeneralize [] (addRegEff (getType tb.sbody)) (newRegVar ()))
	else 
	  (tb.sid, 
	   Normal(addRegEff (getType tb.sbody)))
      in 
      let newEnv = List.map makeNewEnvTbind tblst in
      
      (* ĶΤȤ «ͤη *)
      let rec inferTbind (livreg,liveff) benv (rs:rsubst) (tblst:tbind list) = 
	match (benv,tblst) with
	  ((v,Normal et)::btl,hd::tl) -> 
	    let tmpenv = (substToRenv rs (newEnv@renv)) in
	    let (tmpS,rbody) = inferS (livreg,liveff) tmpenv hd.sbody in
	    let tmpS' = rsubstCombine tmpS rs in
	    let tmpS = 
	      rsubstCombine (unifyRtype (getRtype rbody) (substToType tmpS' et)) tmpS'
	    in 
	    (* ȥåץ٥Υ꡼ unify  *)
	    let rbody = substToRexp tmpS rbody in
	    let topS = topLevel_unify (getRtype rbody) in
	    let tmpS = rsubstCombine topS tmpS  in
	    let tmpS2 = rsubstCombine tmpS rs in
	    let rbody = substToRexp tmpS2 rbody in
	    let (lreg,leff) = fv_of_rexp rbody in
	    (* ɬפΤʤsubstϻĤʤ褦ˤ *)
	    let tmpS2 = subst_filter (livreg, liveff) tmpS2 in
	    let (s_ret,rtl) = inferTbind (lreg@livreg, leff@liveff) btl tmpS2 tl in

	    (s_ret, (substToRexp s_ret rbody)::rtl)
	| ((v,_)::btl,hd::tl) -> 
	    let tmpenv = (substToRenv rs (newEnv@renv)) in
	    let (tmpS,rbody) = inferS (livreg,liveff) tmpenv hd.sbody in
	    let tmpS' = rsubstCombine tmpS rs in
	    (* ȥåץ٥Υ꡼ unify  *)
	    let topS = topLevel_unify (getRtype rbody) in
	    let tmpS = rsubstCombine tmpS topS in
	    let (lreg,leff) = fv (getRtype (substToRexp tmpS rbody)) in
	    let (s_ret,rtl) = inferTbind (lreg@livreg,leff@liveff) btl (rsubstCombine tmpS rs) tl in
	    (s_ret, (substToRexp s_ret rbody)::rtl)
	      
	| (_,[]) -> (rs,[])
	| _ -> failwith "never happen"
      in 
      let (bndS,bnds) = inferTbind (Mylist.append2 (fv_of_Env newEnv) (livreg,liveff)) newEnv rsubst_id tblst in
      
      (* «η *)
      let rbinds = 
	List.map 
	  (function ((_,rs),(tb,(re:rexp))) -> 
	    if isRValue re 
	    then {rid=tb.sid; 
		  rt= rgeneralize (substToRenv bndS (newEnv@renv)) 
		    (substToType bndS (getRtype re)) 
		    (match rs with Forall(_,_,_,_,reg) -> reg | _ -> failwith "AlgorightS.error");
		  rbody = re; (* ޤȤRvarѤ *)
		}
	    else {rid=tb.sid; 
		  rt= Normal (substToType bndS (getRtype re));
		  rbody = re; (* ޤȤRvarѤ *)
		})
	  (List.combine newEnv (List.combine tblst bnds))
      in	
      
      let replaces = 
	List.map (function (rbd,(x,oldscm)) -> replaceRscheme rbd.rid oldscm rbd.rt)
	  (List.combine rbinds newEnv) in
      let replaced_rbinds = 
	List.map (function rbd -> {rid = rbd.rid;
				   rt=rbd.rt;
				   rbody= 
				   List.fold_right (fun x y -> x y)
				     replaces rbd.rbody;
				 }) rbinds
      in
      let needvars = 
	List.fold_right Mylist.append2 
	  (List.map (function b -> fv_of_rexp b.rbody) replaced_rbinds) (livreg,liveff)  in
      let infedEnv = List.map (function x -> (x.rid,x.rt)) replaced_rbinds in
      let (bodyS, rbody) = inferS needvars (substToRenv bndS (infedEnv@renv)) te in
      let getEffect_of_rbind (rbnd:rbind) : effect = 
	match rbnd.rt with
	  Forall(_,_,_,_,r) -> [Reg(r) ]
	| Normal(rt) -> getEffect rbnd.rbody
      in 
      let letrecS = rsubstCombine bodyS bndS in
      (letrecS,
       insertLetRegion 
	 (substToRenv letrecS renv)
	 (RLetrec(List.map (function x-> {rid=x.rid;
					  rt= substToRScheme bodyS x.rt;
					  rbody=substToRexp bodyS x.rbody;
					}) 
		    replaced_rbinds, 
		  rbody,
		  getRtype rbody,
		  List.fold_right (@) 
		    (List.map getEffect_of_rbind replaced_rbinds)
	            (getEffect rbody))))
  | TIf(t1,t2,t3,st) -> debug "if";
      let (s1,r1) = inferS (livreg,liveff) renv t1 in
      let live1 = Mylist.append2 (fv_of_rexp r1) (livreg,liveff) in
      let (s2,r2) = inferS live1 (substToRenv s1 renv) t2 in
      let s2' = rsubstCombine s2 s1 in
      let live2 = Mylist.append2 (fv_of_rexp r2) live1 in
      let (s3,r3) = inferS live2 (substToRenv s2' renv) t3 in
      let s3' = rsubstCombine s3 s2' in
      let su = unifyRtype (substToType s3' (getRtype r2)) (getRtype r3) in
      let sr =  rsubstCombine su s3' in
      let (r1',r2',r3') = 
	(substToRexp sr r1,substToRexp sr r2,substToRexp sr r3) in
      (sr,
       insertLetRegion 
	 (substToRenv sr renv) 
	 (RIf(r1',r2',r3',
	      getRtype r2',
	      sortEffect ((getEffect r1')@(getEffect r2')@(getEffect r3')))))
