open Rtype
open KNormal
open RSyntax

type ident = string
type label = string
type arg_format = (ident list * (ident * region) option)

(* closure$BJQ498e$N<0(B *)
type cexp =
    CVal of rvalue
  | CVar of ident
  | CLet of cbind * cexp
  | CMake_Closure of  ident * (label * ident list * region list * region) * cexp
  | CMake_RClosures of make_rclosure list * cexp 
  | CApply_Closure of ident * ident list
  | CRApply_Closure of ident * region list
  | CIf of ident * cexp * cexp
  | CLetregion of region list * cexp
and cbind =
    { cid: ident ; cbody: cexp }
and make_rclosure = 
    RClosure of ident * (label * ident list * region list * region)
  | RCVal of ident * cexp * region
      
(* closure $BDj5A(B *)
and closure_def = 
    { label: ident;
      fvars: ident list;
      frvars:region list;
      args:  arg_format;
      body:  cexp;
      reg: region;
    }
(* region closure $BDj5A(B *)
and rclosure_def = 
    { r_label: ident;
      r_fvars: ident list;
      r_frvars:region list;
      r_rargs: region list;
      r_body:  cexp;
      r_reg:     region
    }
(* closure $BJQ498e$N(B $B%W%m%0%i%`(B *)
type cprogram = (closure_def list) * (rclosure_def list) * cexp 
      
let new_name = Syntax.new_name

let tolabel ident = "__func__" ^ ident

(* $B<+M3JQ?t$H<+M3%j!<%8%g%sJQ?t$r8+$D$1$F$/$k(B *)
let rec fv_of_cexp ce :(ident list * region list) = 
  match ce with
    CVal(RNumber(_,r)) -> ([],[r])
  | CVal(RBoolean(_))  -> ([],[])
  | CVal(RCharacter(_,r))  -> ([],[r])
  | CVal(RString(_,r))  -> ([],[r])
  | CVal(RSymbol(_,r))  -> ([],[r])
  | CVal(RNull)  -> ([],[])
  | CVal(RUndef)  -> ([],[])
  | CVar(v) -> ([v],[])
  | CLet(bind,body) ->
      let (bfv,bfrv) = fv_of_cexp bind.cbody in
      let (fv,frv) = fv_of_cexp body in
      (Mylist.subtract_list (Mylist.uniq (bfv@fv)) [bind.cid],
       Mylist.uniq (bfrv@frv))
  | CMake_Closure(var,(label,idlst,regs,r),body) ->
      let (fvs,frvs) = fv_of_cexp body in
      (Mylist.uniq (idlst@(Mylist.subtract_list fvs [var])),
       Mylist.uniq (r::regs@frvs))
  | CMake_RClosures(mclos,body) ->
      let fv_of_mkclo = function 
	  RClosure(var,(label,idlst,reglist,reg)) -> 
	    (var::idlst,reg::reglist)
	| RCVal(id,body,reg)  -> 
	    let (fv,frv) = fv_of_cexp body in
	    (fv,reg::frv)
      in 
      let (bfv,bfrv) = 
	List.fold_right Mylist.append2 (List.map fv_of_mkclo mclos) ([],[]) in
      let (fv,frv) = fv_of_cexp body in
      let bind_vars = 
	List.map (function RClosure(var,(id,_,_,_)) -> var | RCVal(id,_,_) -> id) mclos in
      (Mylist.subtract_list (bfv@fv) bind_vars,
       bfrv@frv)
  | CApply_Closure(f,arg) -> (Mylist.uniq (f::arg),[])
  | CRApply_Closure(f,regs) -> ([f],Mylist.uniq regs)
  | CIf(id,e2,e3) ->
      let (v2,r2) = fv_of_cexp e2 in
      let (v3,r3) = fv_of_cexp e3 in
      (Mylist.uniq (id::v2@v3),
       Mylist.uniq (r2@r3))
  | CLetregion(regs,body) ->
      let (fv,frv) = fv_of_cexp body in
      (fv,Mylist.subtract_list frv regs)

let rec kexp2cprg (ke:kexp) : cprogram =
  match ke with
    KVal v ->  ([],[],CVal v)
  | KVar id -> ([],[],CVar id)
  | KFApp(v,regs) ->
      ([],[],CRApply_Closure(v,regs))
  | KLLambda(fname,args,kbody,reg) ->
      let (defs,rdefs,cbody) = kexp2cprg kbody in
      let (fvars, frvars) = fv_of_cexp cbody in
      let fvars' = Mylist.subtract_list fvars (args_to_list args) in
      let new_closure = {label=tolabel fname; 
			 fvars=fvars';
			 frvars=frvars;
			 args=args;
			 body=cbody;
			 reg = reg
		       }
      in (new_closure::defs,rdefs,
	  CMake_Closure(fname,(tolabel fname,new_closure.fvars,new_closure.frvars,reg),
			CVar(fname)))
  | KApp(id,idlst) -> ([],[],CApply_Closure(id,idlst))
  | KLet(bind,kbody) -> 
      let (defs1,rdefs1,cbody1) = kexp2cprg bind.kbody in
      let (defs2,rdefs2,cbody2) = kexp2cprg kbody in
      (defs1@defs2,  rdefs1@rdefs2,
       CLet({cid=bind.kid; cbody=cbody1},
	    cbody2))
  | KLetrec(recbinds, body) ->
      let kbind_to_cbind b = 
	match b with
	  K_Regfun(fname,regs,kbody,reg) ->
	    let (defs, rdefs, cbody) = kexp2cprg kbody in
	    let (fvars,frvars) = fv_of_cexp cbody in
	    let frvars' = Mylist.subtract_list frvars regs in
	    let new_rdef = {r_rargs=regs;
			    r_label=tolabel fname;
			    r_fvars=fvars;
			    r_frvars=frvars';
			    r_body=cbody;
			    r_reg = reg
			  }
	    in 
	    (defs,  new_rdef::rdefs, 
	     RClosure(fname,(tolabel fname,fvars,frvars',reg)))
	| K_VBind(vname,kbody,reg) ->
	    failwith "Error at Closure.ml : The value not a function is binded by letrec syntax."
            (* let (defs, rdefs, cbody) = kexp2cprg kbody in 
	    (defs,rdefs,RCVal(vname,cbody,reg)) *)
      in 
      let (bdefs,brdefs,mclos) = 
	List.fold_right 
	  (fun (defs,rdefs,b) (alldefs,allrdefs,bs) -> 
	    (defs@alldefs,  rdefs@allrdefs,  b::bs))
	(List.map kbind_to_cbind recbinds) ([],[],[]) 
      in 
      let (defs, rdefs, cbody) = kexp2cprg body in
      (bdefs@defs,  brdefs@rdefs, 
       CMake_RClosures(mclos,cbody))
  | KIf(id,k2,k3) ->      let (defs2, rdefs2, cbody2) = kexp2cprg k2 in
      let (defs3, rdefs3, cbody3) = kexp2cprg k3 in
      (defs2@defs3,  rdefs2@rdefs3, 
       CIf(id,cbody2,cbody3))
  | KLetregion(regs,kbody) ->
      let (defs, rdefs, cbody) = kexp2cprg kbody in
      (defs,    rdefs, 
       CLetregion(regs,cbody))




(* ============= pretty printing ==============*)

open Format

let rec comma_sep strlst = 
  match strlst with 
    [] -> ""
  | [str] -> str
  | hd::tl -> hd ^ "," ^ (comma_sep tl)

let rec arg_to_str arg= 
  match arg with
    (args,Some(o,r)) -> (comma_sep args) ^ ". (" ^ o ^ "," ^ (string_of_reg r) ^ ")"
  | (args,None) -> (comma_sep args) 

let print_cprogram (clolst, rclolst, main) =
  let rec pr_cexp ppf cexp = 
    match cexp with
      CVal(v) -> fprintf ppf "%s" (string_of_rvalue v)
    | CVar(v) -> fprintf ppf "%s" v
    | CLet(bind,body) -> 
	fprintf ppf "@[<hov 0>let@ %s@ =@ %a@ @,in@ @,%a@]" 
	  bind.cid pr_cexp bind.cbody pr_cexp body
    | CMake_Closure(var,(id,ids,regs,reg),body) ->
	fprintf ppf "@[<hov 0>let %s = make_closure (Label:%s(%s)(%s)@ at@ %s@ @,in@ @,%a@]" 
	  var id (comma_sep ids) (comma_sep (List.map string_of_reg regs)) 
	  (string_of_reg reg) pr_cexp body
    | CMake_RClosures(mkclos, body) ->
	fprintf ppf "@[<hov 0>%a@ @,in@ @,%a@]" 
	  pr_mkrclos mkclos pr_cexp body
    | CApply_Closure(id,ids) ->
	fprintf ppf "@[<hov 0>apply_closure(%s@ :@ %s)@]" 
	  id (comma_sep ids) 
    | CRApply_Closure(id,regs) ->
	fprintf ppf "@[<hov 0>Rapply_closure(%s@ :@ %s)@]" 
	  id (comma_sep (List.map string_of_reg regs)) 
    | CIf(id,c1,c2) ->
	fprintf ppf "@[<hov 0>if(%s)@,{%a}@,else{%a}@]" 
	  id pr_cexp c1 pr_cexp c2
    | CLetregion(regs,body) ->
	fprintf ppf "@[<hov 0>letregion(%s)@,{%a}@]" 
	  (comma_sep (List.map string_of_reg regs)) pr_cexp body
  and pr_mkrclos ppf clos = 
    match clos with
      RClosure(var,(label,ids,regs,reg))::tl ->
	fprintf ppf "@[<hov 0>let %s@ =@ make_rclosure@ (Label:%s(%s)(%s),%s)@\n%a@]" 
	  var label (comma_sep ids) (comma_sep (List.map string_of_reg regs)) 
	  (string_of_reg reg) pr_mkrclos tl
    | RCVal(id,cexp,reg)::tl ->
	fprintf ppf "@[<hov 0>let %s@ at@ %s@ =@, %a@\n%a@]" 
	  id (string_of_reg reg) pr_cexp cexp pr_mkrclos tl
    | [] -> fprintf ppf ""
  in 
  let rec pr_clo_lst ppf clos = 
    match clos with 
      clo::tl ->
	fprintf ppf "@[<hov 0>%s : {(%s)(%s)}(%s)@\n\t%a@\n%a@]" 
	  clo.label (comma_sep clo.fvars) 
	  (comma_sep (List.map string_of_reg clo.frvars)) (arg_to_str clo.args) 
	  pr_cexp clo.body  pr_clo_lst tl
    | [] -> fprintf ppf ""
  in 
  let rec pr_rclo_lst ppf rclos = 
    match rclos with
      rclo::tl ->
	fprintf ppf "@[<hov 0>%s : {(%s)(%s)}(%s)@\n\t%a@\n%a@]" 
	  rclo.r_label (comma_sep rclo.r_fvars) 
	  (comma_sep (List.map string_of_reg rclo.r_frvars))
	  (comma_sep (List.map string_of_reg rclo.r_rargs))
	  pr_cexp rclo.r_body  pr_rclo_lst tl
    | [] ->  fprintf ppf ""
  in 
  printf "@[<hov 0>%a@\n%a@\n%a@]" pr_clo_lst clolst pr_rclo_lst rclolst pr_cexp main
