
(* let _ = print_endline "debug.ml" *)

let parse_string str = 
  Parser.program Lexer.token (Lexing.from_string str) 


let parse_file filename =
  let ic =  open_in filename in 
  let r = 
    try
      Syntax.remove_set (Syntax.program2exp (Parser.program Lexer.token (Lexing.from_channel ic)))
    with 
      Parsing.Parse_error ->
	begin 
	  print_endline ("parse error : line " ^ (string_of_int !(Lexer.line)));
	  raise Parsing.Parse_error
	end
  in
  close_in ic; r

let typing (e:Syntax.exp) : (Stype.subst * TSyntax.texp) = TSyntax.stype_infer [] e


let open_graph filename title = 
  Unix.system ("sh test.gpl "^ filename ^ " " ^ title)
    
(*
let debug fname =
  Stype.print_stype (TSyntax.getType (snd (typing (parse_file fname))))

let debug_str str =
  Stype.print_stype (TSyntax.getType (snd (typing (parse_string str))))
*)
  
let _ =
  if not !Sys.interactive
  then
    try 
      begin
	
	let exp = parse_file Sys.argv.(1) in

	(* alpha Ѵ *)
	let predefined_funcs = Builtins.predefined_variables in
	let aexp = Syntax.alpha_conv 
	    (List.combine predefined_funcs predefined_funcs)
	    exp in
	
	let (st,texp) = TSyntax.stype_infer Builtins.builtin_senv aexp in
	
	(* ꡼ *)
	AlgorithmS.size := TSyntax.size texp;
	let (rsub,rexp) = AlgorithmS.inferS ([],[]) Builtins.builtin_renv texp in 
	let (rsub2,rexp2) = AlgorithmR.inferR Builtins.builtin_renv rexp in
	let r2exp = R2Syntax.allocates_global_regions (R2Syntax.rexp_to_r2exp rexp2) in
	R2Syntax.simple_print_r2exp r2exp;  
	flush_all ();
	
      end
    with 
      Failure(st) -> print_string (st ^ "\n\n")
    | e -> raise e 
  else 
    ()
    

let read_and_typing file = 
  let exp = parse_file file in
  let predefined_funcs = Builtins.predefined_variables in
  let aexp = Syntax.alpha_conv 
      (List.combine predefined_funcs predefined_funcs)
      exp in
  let (st,texp) = TSyntax.stype_infer Builtins.builtin_senv aexp in
  let (rsub,rexp) = AlgorithmS.inferS ([],[]) Builtins.builtin_renv texp in
  let (rsub2,rexp2) = AlgorithmR.inferR Builtins.builtin_renv rexp in
  rexp2
  (* RSyntax.allocates_global_regions rexp2 *)

let exec e =
  Interp.interp2 Builtins.builtin_venv (R2Syntax.allocates_global_regions (R2Syntax.rexp_to_r2exp e))

let exec_file file = 
  exec (read_and_typing file)
