(define (int->float x) x)
(define float->int truncate)

(define CR #\newline)
(define SPACE #\space)

(define (call-with-input-file filename proc)
  (let ((inport (open-input-file filename)))
    (begin (proc inport)
	   (close-input-port inport))))
(define (call-with-output-file filename proc)
  (let ((outport (open-output-file filename)))
    (begin (proc outport)
	   (close-output-port outport))))

(define (caar p)
  (car (car p)))
(define (cdar p)
  (cdr (car p)))


(define (fsqr x) (* x x))

(define (!= a b) (not (= a b)))
(define f!= !=)

(define (sgn x) 
  (cond ((> x 0.0)  1.0)
	((< x 0.0) -1.0)
	(else x)))


(define object (make-vector 60))


(define (vector-initialize v num max)
  (if (< num max)
      (begin (vector-set! v num 0)
	     (vector-initialize v (+ 1 num) max))
      '()))

(define (set-object i)
  (if (< i 60)
      (let ((v  (make-vector 20)))
	(begin
	  (vector-initialize v 0 20)
	  (vector-set! object i v)
	  (set-object (+ i 1)))
	)))

; size = (x$B<4$NAv::@~K\?t(B . y$B<4$NAv::@~K\?t(B)

(define size (cons 128 128))

;$B<B9T;~%*%W%7%g%s(B

(define dbg (list #f))
(define sh  (list #t))

; Screen$B$N:BI8(B
(define Screen (make-vector 3))

;$B;kE@$N:BI8!J(Bofs$B$J$7!K(B
(define Vp     (make-vector 3))

;$B;kE@$N:BI8(B (ofs$B$"$j(B)
(define View   (make-vector 3))

;$B8w@~%Y%/%H%k!JC10L%Y%/%H%k!K!a8w$N8~$-(B
(define Light  (make-vector 3))

;$B=hM}$K;HMQ$9$k!#(BcosV1,sinV1,cosV2,sinV2$B$r3JG<(B

;(cosV1 . cosV2)
(define cos-V (cons 0 0))

;(sinV1 . sinV2)
(define sin-V (cons 0 0))

;$B6@LL$N%O%$%i%$%H6/EY(B $B!JI8=`$G(B 255$B!K(B
(define beam (list 0))

;AND$B%M%C%H%o!<%/(B $B$N%G!<%?$rF~$l$k%Y%/%?!J:GBg(B50$B8D!K(B

(define and-Net (make-vector 50))

;OR$B%M%C%H%o!<%/(B
(define or-Net (list 0))


;x /180.0 * pi(= 3.1415927)

(define (rad x) (* x 0.017453293))

(define (Read-Parameter InputFile)

  (call-with-input-file InputFile
    (lambda (port)
      
      (begin

	;$B%9%/%j!<%s!"8w8;$J$I$N%G!<%?$NFI$_9~$_(B
	(Read-Environ port)

	;$B%*%V%8%'%/%H$N%G!<%?$NFI$_9~$_(B
	(Read-All-Object port)
  
	;AND$B%M%C%H%o!<%/$NFI$_9~$_(B
	(Read-And-Network 0 port)
	
	;OR$B%M%C%H%o!<%/$NFI$_9~$_(B
	(set-car! or-Net (Read-Or-Network port))
	
	))))

;--------------------------------------------------------------------
;  Read-Environ : $B%9%/%j!<%s$K4X$9$k%Q%i%a!<%?$rFI$_9~$`(B
;--------------------------------------------------------------------

(define (Read-Environ port)
  (begin

					;$B%9%/%j!<%sCf?4$N:BI8$r%;%C%H(B
    (vector-set! Screen 0 (read port))
    (vector-set! Screen 1 (read port))
    (vector-set! Screen 2 (read port))

    (let* ((v1  (rad (read port)))
	   (cv1 (cos v1))
	   (sv1 (sin v1))
	   
	   (v2  (rad (read port)))
	   (cv2 (cos v2))
	   (sv2 (sin v2))
	   
	   (nl  (read port))
	   
	   (l1  (rad (read port)))
	   (cl1 (cos l1))
	   (sl1 (sin l1))
	   
	   (l2  (rad (read port)))
	   (cl2 (cos l2))
	   (sl2 (sin l2)))

      (begin
	(set-car! beam (read port))
	
	;$B8w@~%Y%/%H%k$N%;%C%H(B
	(vector-set! Light 0 (* cl1 sl2))
	(vector-set! Light 1 (- sl1))
	(vector-set! Light 2 (* cl1 cl2))

	;$B;kE@0LCV%Y%/%H%k!J%m!<%+%k:BI8!K$N%;%C%H(B
	(vector-set! Vp 0 (* cv1 sv2  -200.0))
	(vector-set! Vp 1 (* (- sv1) -200.0))
	(vector-set! Vp 2 (* cv1 cv2  -200.0))

	;$B=hM}$K;HMQ$9$k(B cos,sin $B$NMQ0U(B
	(set-car! cos-V cv1)
	(set-car! sin-V sv1)
	;(write sin-V)(newline)
	(set-cdr! cos-V cv2)
	(set-cdr! sin-V sv2)

	;$B;kE@0LCV%Y%/%H%k!J%o!<%k%I:BI8!K$N%;%C%H(B
	(vector-set! View 0 
		     (+ (vector-ref Vp 0) (vector-ref Screen 0)))
	(vector-set! View 1
		     (+ (vector-ref Vp 1) (vector-ref Screen 1)))
	(vector-set! View 2
		     (+ (vector-ref Vp 2) (vector-ref Screen 2)))

	))))

;--------------------------------------------------------------------
;  Read-Nth-Object : $B%U%!%$%k$+$i(Bn$B8DL\$N%*%V%8%'%/%H$rFI$_9~$`4X?t(B
;--------------------------------------------------------------------

(define (Read-Nth-Object n port)

  (let ((Texture (read port)))
    (if (!= Texture -1) 
	
	(let ((obj (vector-ref object n)))
	   
	  (begin
	    (vector-set! obj 0 Texture)
	    (vector-set! obj 1 (read port))
	    (vector-set! obj 2 (read port))
	    
	    ;xscheme$B$N>l9g(B let$BJ8(B $B$G$N(B $B6I=jJQ?t$X$NBeF~$N:]$K(B
	    ;(read port)$B$,J#?t$"$k$H:G8e$N(B(read port)$B$NCM(B $B$,<h$i$l$k(B
	    ;$B$D$^$j!"(B(read port)$B$,O"B3$GI>2A$5$l$k>l9g!"@h$NCM$OL5;k(B
	    ;$B$5$l$k!#4X?t$X$NBeF~$N:]$K$b5$$r$D$1$k$3$H!#(B

	    (let ((rtset? (if (= (read port) 1) #t #f)))
	      
	      (begin
		(vector-set! obj  3 rtset?)
		(vector-set! obj  4 (read port))
		(vector-set! obj  5 (read port))
		(vector-set! obj  6 (read port))
		(vector-set! obj  7 (read port))
		(vector-set! obj  8 (read port))
		(vector-set! obj  9 (read port))
		(vector-set! obj 10 (read port))
		(vector-set! obj 11 (read port))
		(vector-set! obj 12 (read port))
		(vector-set! obj 13 (read port))
		(vector-set! obj 14 (read port))
		(vector-set! obj 15 (read port))
		
		;$B!}(B $B%*%V%8%'%/%H$r2sE>$5$;$k;~(B
		(if rtset?
		    (begin
		      (vector-set! obj 16 (read port))
		      (vector-set! obj 17 (read port))
		      (vector-set! obj 18 (read port))
		      ))
		
		#t))))
	#f)))

;--------------------------------------------------------------------
;  Modify-Nth-Object 
;
;    $BFI$_9~$s$@%*%V%8%'%/%H$N%G!<%?$r(B
;           $BJ?LL$N>l9g!"(Ba^2 + b^2 + c^2 = 1 $B$K$J$k$h$&$K=$@5(B
;           $B#2<!6JLL$N>l9g(B x^2/a + y^2/b + Z^2/c = 1$B$H$J$k$h$&$K=$@5(B 
;
;    $B%*%V%8%'%/%H$N2sE>$,I,MW$J>l9g$K$O2sE>$r%5%]!<%H$9$k!#(B
;--------------------------------------------------------------------

(define (Modify-Nth-Object n port)

  (let* ((obj    (vector-ref object n))
	 (Form   (vector-ref obj 1))
	 (rtset? (vector-ref obj 3)))
    
    (begin
      
      ;$B!}(B $B%Q%i%a!<%?(BA,B,C$B$OJ*BN$N<oN`$K$h$C$F$OJQ99$NI,MW$"$j(B
      
      ;Form=3 $B%G!<%?$O(B2$B<!6JLL$,(B1$BHVB?$$$+$i@h$K!#(B
      (cond 
       
       ((= Form 3)
	(let ((A (vector-ref obj 4))
	      (B (vector-ref obj 5))
	      (C (vector-ref obj 6)))
	  (begin
	    (vector-set! obj 4 (if (= A 0.0) 0.0
				   (/ (sgn A)(fsqr A))))
	    (vector-set! obj 5 (if (= B 0.0) 0.0
				   (/ (sgn B)(fsqr B))))
	    (vector-set! obj 6 (if (= C 0.0) 0.0
				   (/ (sgn C)(fsqr C))))
	    )))
     
       ;Form=2 $BJ*BN$,J?LL$N>l9g$N=$@5(B
       ((= Form 2)
	(let* ((A (vector-ref obj 4))
	       (B (vector-ref obj 5))
	       (C (vector-ref obj 6))
	       (Q (sqrt (+ (fsqr A)(fsqr B)(fsqr C)))))
	  (if (= (vector-ref obj 10) 1.0)
	      (begin
		(vector-set! obj 10 -1.0)
		(vector-set! obj 4 (/ (- A) Q))
		(vector-set! obj 5 (/ (- B) Q))
		(vector-set! obj 6 (/ (- C) Q)))
	      (begin 
		(vector-set! obj 4 (/ A Q))
		(vector-set! obj 5 (/ B Q))
		(vector-set! obj 6 (/ C Q))
		)))))
           
      ;Form=1 $BJ*BN$,$=$NB>$N;~$O=$@5$J$7!#(B
      
      ;$B!}(B $B%*%V%8%'%/%H$r2sE>$5$;$k;~(B
      (if rtset? 
	  
	  ;$B2sE>3Q$NFI$_9~$_(B 
	  (let* ((RotX (vector-ref obj 16))
		 (RotY (vector-ref obj 17))
		 (RotZ (vector-ref obj 18))
	  			   
		 ;$B=hM}$K;HMQ$9$k(Bcos,sin$B$N:n@.(B
		 (cx (cos (rad RotX)))
		 (sx (sin (rad RotX)))
		 (cy (cos (rad RotY)))
		 (sy (sin (rad RotY)))
		 (cz (cos (rad RotZ)))
		 (sz (sin (rad RotZ)))
		 
		 ;3$B<!852sE>$N9TNs(B
		 (r0 (* cy cz))
		 (r1 (- (* sx sy cz) (* cx sz)))
		 (r2 (+ (* cx sy cz) (* sx sz)))
		 (r3 (* cy sz))
		 (r4 (+ (* sx sy sz) (* cx cz)))
		 (r5 (- (* cx sy sz) (* sx cz)))
		 (r6 (- sy))
		 (r7 (* sx cy))
		 (r8 (* cx cy))
		 
		 (ao (vector-ref obj 4))
		 (bo (vector-ref obj 5))
		 (co (vector-ref obj 6)))
	    
	    (begin
	      
	      ;$B%Q%i%a!<%?$NJQ99(B
	      (vector-set! obj 4 (+ (* ao (fsqr r0))(* bo (fsqr r3))(* co (fsqr r6))))
	      (vector-set! obj 5 (+ (* ao (fsqr r1))(* bo (fsqr r4))(* co (fsqr r7))))
	      (vector-set! obj 6 (+ (* ao (fsqr r2))(* bo (fsqr r5))(* co (fsqr r8))))
	    
	      ;$B2sE>3Q$NJQ99(B
	      (vector-set! obj 16 (* 2.0 (+ (* ao r1 r2) (* bo r4 r5) (* co r7 r8))))
	      (vector-set! obj 17 (* 2.0 (+ (* ao r0 r2) (* bo r3 r5) (* co r6 r8))))
	      (vector-set! obj 18 (* 2.0 (+ (* ao r0 r1) (* bo r3 r4) (* co r6 r7))))
	      ))))))
  

;--------------------------------------------------------------------
; Read-Object
;
; $B%*%V%8%'%/%H$N%G!<%?$rFI$_9~$_!"(B
; $B:G8e$N%*%V%8%'%/%H$NF~$C$?%Y%/%?$NHV9f$rJV$9!#(B
;
;--------------------------------------------------------------------

(define (Read-Object n port)
  (if (< n 61)
      (if (Read-Nth-Object n port)
	  (Read-Object (+ n 1) port)
	  (- n 1))
      ))

;--------------------------------------------------------------------
; Modify-Object : $B%*%V%8%'%/%H$NF~$C$?%Y%/%?$r(B0$B$N%G!<%?$^$G=g$K=$@5(B
;--------------------------------------------------------------------

  (define (Modify-Object n port)
    (if (>= n 0)
	(begin 
	  (Modify-Nth-Object n port)
	  (Modify-Object (- n 1) port))
	#f))

;--------------------------------------------------------------------
; Read-All-Object : $B%*%V%8%'%/%H$N%G!<%?$rFI$_9~$_I,MW$K1~$8=$@5(B
;--------------------------------------------------------------------

(define (Read-All-Object port)
  (Modify-Object (Read-Object 0 port) port))


;--------------------------------------------------------------------
; Read-Network : $B#A#N#D!"#O#R%M%C%H%o!<%/$rFI$_9~$_!"(B
;--------------------------------------------------------------------

;$B%M%C%H%o!<%/Cf$N%"%$%F%`$r(B -1$B$^$G$r=8$a%j%9%H$K$9$k!#(B
(define (Read-Net-Item port)
  (let ((item (read port)))
    ;(write 'item)(write item)(newline)
    (if (= item -1)
	'()
	(cons item (Read-Net-Item port)))))

;OR$B%M%C%H%o!<%/$r=g$KJB$Y$?%j%9%H$r:n$k!#(B
(define (Read-Or-Network port)
  (let ((net (Read-Net-Item port)))
    (if (null? net)
	'()
	(cons net (Read-Or-Network port)))))

;AND$B%M%C%H%o!<%/$r=g$K%Y%/%?(BandNet$B$K3JG<!#(B
(define (Read-And-Network n port)
  (let ((net (Read-Net-Item port)))
    (if (null? net)
	#t
	(begin
	  (vector-set! and-Net n net) 
	  (Read-And-Network (+ n 1) port)))))


;$BE@$r%l%$%H%l!<%9$9$k4X?t(B
;(load "Sca-scm")
;------------------------------------------------------------------------------
;       Ray Tracing Program for chez
;
;    original program by Ryoji Kawamichi
;            arranged by Motohico Nanano
;
;               Scan.scm
;
;         last change : 1996/4/8
;-----------------------------------------------------------------------------

; $B%9%/%j!<%s$NA4E@$N%9%-%c%s$r9T$J$&4X?t72(B
;
;Scan-Start <-$B30It$+$i8F$P$l$k!#(B
;
;Write-PPM-Header
;Write-RGB
;Scan-Line
;Scan-Point
;RayTracing
;
;$B;kE@$+$i%9%/%j!<%s>e$NA4E@$KBP$7!"%l%$%H%l!<%9$r;O$a$k!#(B
; y$B<4$NAv::@~$N%*%U%;%C%H$O(BBASIC$B$H0c$$<B9T;~%Q%i%a!<%?$K4^$^$l$F$$$J(B
;$B$$$?$a(B 
;
;  x$B<4(B 1 $B$+$i(B SizeX $B$KBP1~$9$kAv::@~>e$rD4$Y$k!#(B
;  y$B<4(B 1 $B$+$i(B SizeY $B$KBP1~$9$kAv::@~>e$rD4$Y$k!#(B 


;-------------------------------------------------------------------
; Scan-Start :
;  $B!&=PNO%U%!%$%k$r%*!<%W%s(B
;  $B!&(BPPM$B%U%)!<%^%C%H$GI,MW$J(B Header $B$r=q$-9~$`!#(B
;  $B!&(BScan-Line$B$r:F5"E*$KMQ$$(By$B<4$NA4Av::@~>e$rD4$Y$k!#(B
;  $B!&=PNO%U%!%$%k$r%/%m!<%:$9$k!#(B
;-------------------------------------------------------------------






;-------------------------------------------------------------------
;  rt : $B%a%$%s4X?t(B
;
;  $B!&(B $B<B9T;~$N%Q%i%a!<%?$N3JG<!#(B
;  $B!&(B $B%G!<%?$rFI$_9~$`(B
;  $B!&(B $B%9%/%j!<%s$N3FE@$K$D$$$F%l%$%H%l!<%9$9$k!#(B
;  $B$H$$$C$?A`:n$r9T$J$&!#(B
;
;  InputFileName  : $B%U%)!<%^%C%H$K$7$?$,$C$?F~NO%G!<%?(B
;  OutputFileName : $B=PNO%U%!%$%kL>(B
;  size-x         : x$B<4J}8~$NAv::@~K\?t(B
;  size-y         : y$B<4J}8~$NAv::@~K\?t(B
;  Debug?         : $B:#!"(By$B<4J}8~$NAv::@~$N2?K\L\$rD4$Y$F$$$k$+I=<($9$k$+!)(B
;  Shadow?        : $B1F$r$D$1$k$+!)(B
;-------------------------------------------------------------------
(define (showlst lst)
  (if (null? lst)
      '()
      (begin ;(write (car lst))(newline)
	     (showlst (cdr lst)))))

(define (showornet ornet)
  (if (null? ornet)
      '()
      (begin (showlst (car ornet))
	     (showornet (cdr ornet)))))
(define (showornet2 ornet)
  (if (null? ornet)
      '()
      (begin (showornet (car ornet))
	     (showornet2 (cdr ornet)))))

(define (rt InputFileName OutputFileName size-x size-y Debug? Shadow?)
  (begin

    ;CompilerCheck.scm$B$N;HMQ;~$N$_!"3F?t$N=i4|2=(B
    ;(SetN)
    
    ;object$B$NG[Ns$NMQ0U(B
    (set-object 0)
    ;$B<B9T;~%Q%i%a!<%?$NFI$_9~$_(B
    (set-car! size size-x)
    (set-cdr! size size-y)
    (set-car! dbg Debug?)
    (set-car! sh Shadow?)
    
    ;$B%G!<%?$NFI$_9~$_(B
    (Read-Parameter InputFileName)
    (write "readdataok")
    ;(showornet2 or-Net)
    ;(write or-Net)
    ;$B%l%$%H%l!<%9$9$k!#(B
    ; (Scan-Start OutputFileName)
    ; (write "raytre ok")
    ))
;$BF~NO%U%!%$%k(B
(define in "test.sld")
;$B=PNO%U%!%$%k(B  : xv$B$G8+$l$k!#(B
(define out "q5.ppm")
;load$B;~$K<+F0E*$K<B9T$9$k!#(B
(begin
  (rt in out 128 128 #f #t)
  ;(display out)
  ;(exit)
  )

