(* 󥿥ץ꥿ *)

open RSyntax  
open R2Syntax
open Library

let size_of_num = 8
let size_of_string n = n
let size_of_symbol = 4
let size_of_closure = 32
let size_of_vector n = 4*n + 4
let size_of_port = 4
let size_of_pair = 12


(* ¹ԻΥ꡼ *)
module VReg =
  struct 
    type t = (string * int option) ref
	  
    let memory_size = ref 0
    let total_size = ref 0
    let inc_mem i = 
      begin 
	memory_size := (!memory_size + i);
	total_size := (!total_size + i)
      end
    let dec_mem i = memory_size := (!memory_size - i)

    let f = ref None 
    let file_init fname = f := Some(open_out fname)
    let file_fin fname = 
      match !f with
	Some c -> close_out c
      | None -> ()


    let pc = ref 0
    let out_memsize () = 
      match !f with
	Some c ->
	  begin 
	    (if (!pc mod 2) = 0
	    then (Printf.fprintf c "%d\t%d\n" !pc !memory_size)
	    else ());
	    pc := !pc + 1;
	    flush c
	  end
      | None -> ()
    let mark () =
      match !f with
	Some c ->
	  begin
	    Printf.fprintf c "%d\t%d\t0\n" !pc !memory_size;
	    Printf.fprintf c "%d\t%d\t100\n" !pc 0;
	    Printf.fprintf c "%d\t%d\t0\n" !pc !memory_size;
	  end
      | None -> ()

    let regname r : string = 
      match !r with 
	(r,_) -> r
	
    (* ꡼ *)
    let newreg r : t= 
      (* print_endline ("allocate " ^ r);*)
      ref (Rtype.string_of_reg r,(Some(0)))

    (* ꡼Υ֤ *)
    let size (r:t) : int = 
      match !r with 
	(r,Some(s)) -> s
      | (r,None)    -> failwith ("Invalid access to deallocated region(size) " ^ r)

    (* ꡼ *)
    let del (r:t) = 
      match !r with 
	(n,Some(s)) -> 
	    begin 
	      (* print_endline ("deallocate " ^ n); *)
	      dec_mem (size r);
	      r := (n,None);
	      out_memsize ()
	    end
      | (n,None)    -> failwith ("Invalid access to deallocated region(del) " ^ n)
	
    (* ꡼Υ䤹 *)
    let inc (r:t) i = 
      match !r with 
	(n,Some(s)) -> 
	  begin 
	    r:= (n,Some(s + i));
	    inc_mem i;
	    out_memsize ()
	  end
      | (n,None)    -> failwith ("Invalid access to deallocated region(int)." ^ n)
    (* ꡼˥ *)
    let touch (r:t) : unit = 
      match !r with 
	(n,Some(_)) -> ()
      | (n,None)    -> failwith ("Invalid access to deallocated region(touch)." ^ n)
  end

(** read ؿ *)
let (reader_reg: VReg.t option ref) = ref None
let read_get_reg () = 
  match !reader_reg with
    Some s -> s
  | None -> failwith "region is empty"
let read_set_reg r = reader_reg := Some r



(** ѿ̾ *)
type var    = string 
(** ꡼ѿ̾ *)
type regvar = string

(** ¹Ի *)
type scmval = 
    VNum of scmnum * VReg.t
  | VBool of bool 
  | VSymbol of string * VReg.t
  | VString of string * VReg.t
  | VClosure of venv * (ident list * (ident * Rtype.region) option) 
	* r2exp  * VReg.t 
  | VRClosure of venv * Rtype.region list * r2exp  * VReg.t 
  | VPair of c_cell * VReg.t 
  | VChar of char * VReg.t
  | VNill
  | VUndef
  | Vinport of Lexing.lexbuf * in_channel * VReg.t
  | Voutport of out_channel * VReg.t
  | VVector of scmval array * VReg.t
  | VBClosure of (VReg.t list -> scmval) * VReg.t
  | VBuiltin of (scmval list -> scmval) * VReg.t
(**  *)
and c_cell = { mutable car: scmval;  mutable cdr: scmval }

(** ʹĶ *)
and val_env = (var * scmval ref) list
(** ꡼Ķ *)
and vreg_env = (Rtype.region * VReg.t) list
(** Ķ *)
and venv = vreg_env * val_env 

let getVreg regvar regenv  = 
  if List.mem_assoc regvar regenv 
  then List.assoc regvar regenv
  else 
    begin
      (* List.iter (function (x,_) -> print_endline x) regenv; *)
      failwith ((Rtype.string_of_reg regvar) ^ " not found")
    end
let getVal var valenv  = 
  !(if List.mem_assoc var valenv
  then List.assoc var valenv
  else failwith (var ^ " not found"))



let rec string_of_scmval sval = 
  match sval with
    VNum(i,_) -> string_of_scmnum i
  | VBool(true) -> "#t" 
  | VBool(false) -> "#f" 
  | VChar(c,_) -> "#\\" ^ (Char.escaped c)
  | VSymbol(sym,r) -> "'" ^ sym 
  | VString(str,r) -> "\"" ^ str ^ "\""
  | VClosure _ -> "*procedure*"
  | VRClosure _ -> "*procedure*"
  | VPair(v,_) -> "(" ^ (string_of_scmval v.car) ^ " . " ^ (string_of_scmval v.cdr) ^ ")"
  | VNill -> "'()"
  | VVector (v,_) -> "#(" ^ (Array.fold_right (^) (Array.map string_of_scmval v) "") ^ ")"
  | VUndef -> "*undef*"
  | Vinport _ -> "*inport*"
  | Voutport _ -> "*outport*"
  | VBClosure _ -> "*procedure*"
  | VBuiltin _ -> "*procedure*"



let rec interp2 (((vreg_env:(Rtype.region * VReg.t) list) ,val_env) as venv:venv) (e:r2exp) : scmval = 
  try
    match e with
      R2_Val(RNumber(n, regvar)) -> 
	let r = getVreg regvar vreg_env in
	VReg.inc r 4;
	VNum(n,r)
    | R2_Val(RBoolean(true)) -> VBool(true)
    | R2_Val(RBoolean(false)) -> VBool(false)
    | R2_Val(RNull) -> VNill
    | R2_Val(RUndef) -> VUndef
    | R2_Val(RSymbol(str,regvar)) -> 
	let r = getVreg regvar vreg_env in
	VReg.inc r 4;
	VSymbol(str,r)
    | R2_Val(RCharacter(c, regvar)) -> 
	VChar(c,getVreg regvar vreg_env)
    | R2_Val(RString(str, regvar)) -> 
	VString(str,getVreg regvar vreg_env)
(*    | R2_Val(rv) -> 
	print_endline (string_of_rvalue rv);
	failwith "not impl"*)
    | R2_Var(var) -> 
	getVal var val_env 
    | R2_FApp(rf,regs) ->
	begin 
	  match (interp2 venv rf) with
	    VRClosure((c_vreg_env,c_val_env) as venv, regvars, rexp, reg) ->
	      let new_vreg_env = 
		List.combine regvars 
		  (List.map (function r -> getVreg r vreg_env) regs) in
	      interp2 (new_vreg_env@c_vreg_env, c_val_env) rexp
	  | VBClosure(rf,reg) ->
	      rf (List.map (function r -> getVreg r vreg_env) regs)
	  | v -> 
	      failwith ("cannot apply these regions to this value : " 
			^ string_of_scmval v)
	end
    | R2_Lambda(args, rexp, regvar) ->
	VClosure(venv, args, rexp, getVreg regvar vreg_env)
    | R2_Reglambda(regargs, body, regvar) ->
	VRClosure(venv, regargs, body, getVreg regvar vreg_env)
    | R2_App(f,args) ->
	begin
	  match interp2 venv f with
	    VClosure((vreg_env,val_env), (fargs,Some(ftail,ftreg)), fexp, creg) -> 
	      let vargs = 
		List.map 
		  (function x -> ref (interp2 venv x)) args in
	      let rec getNewEnv fargs vargs =
		begin 
		  match (fargs,vargs) with
		    (fhd::ftl, vhd::vtl) ->
		      (fhd,vhd)::(getNewEnv ftl vtl) 
		  | (fhd::ftl, []) ->
		      failwith "argument is not supplyed"
		  | ([],vtl) ->
		      [(ftail,
			let listreg = getVreg ftreg vreg_env in
			ref (List.fold_right 
			       (fun x y -> VPair({car= !x; cdr=y},listreg))
			       vtl VNill))]
		end
	      in 
	      let newenv = getNewEnv fargs vargs in
	      interp2 (vreg_env,newenv@val_env) fexp
	  | VClosure((reg_env,val_env), (fargs,None), fexp, creg) -> 
	      let vargs = 
		List.map 
		  (function x -> ref (interp2 venv x)) args in
	      let rec getNewEnv fargs vargs =
		begin 
		  match (fargs,vargs) with
		    (fhd::ftl, vhd::vtl) ->
		      (fhd,vhd)::(getNewEnv ftl vtl) 
		  | (fhd::ftl, []) ->
		      failwith "argument is not supplyed"
		  | ([],vhd::vtl) ->
		      failwith "argument is too many"
		  | ([],[]) -> []
		end
	      in 
	      let newenv = getNewEnv fargs vargs in
	      interp2 (reg_env,newenv@val_env) fexp
		
	  | VBuiltin(bf,reg) ->
	      (* print_string "exec buildin : "; simple_print_r2exp e; *)
	      let vargs = 
		List.map 
		  (function x -> (interp2 venv x)) args in
	      bf vargs
	  | _ -> failwith "cannot apply values to this value"
	end
    | R2_Let(rblst, rexp) ->
	let newenv = 
	  List.map 
	    (function (id,body) -> (id,
				    ref (interp2 venv body))) rblst in
	interp2 (vreg_env, newenv@val_env) rexp
    | R2_Letrec(rblst, rexp) -> 
	let newenv = List.map 
	    (function 
		RB_Regfun(id,regs,body,r) -> (id,ref VNill)
	      | RB_VBind(id,body,r)       -> (id,ref VNill)) rblst in
	let rec evalbinds binds = 
	  begin
	    match binds with
	      RB_Regfun(id,regs,body,r)::tl -> 
		let v = 
		  interp2 (vreg_env,newenv@val_env) (R2_Reglambda(regs,body,r)) in
		(List.assoc id newenv) := v;
		evalbinds tl
	    | RB_VBind(id,body,r)::tl -> 
		let v = 
		  interp2 (vreg_env,newenv@val_env) body in
		(List.assoc id newenv) := v;
		evalbinds tl
	    | [] -> () 
	  end
	in 
	evalbinds rblst;
	interp2 (vreg_env,newenv@val_env) rexp
    | R2_If(e1, e2, e3) ->
	begin
	  match (interp2 venv e1) with
	    VBool(false) -> 
	      interp2 venv e3
	  | _ -> 
	      interp2 venv e2	  
	end
    | R2_Letregion(regs, rexp) ->
	let (new_vreg_env:(Rtype.region * VReg.t) list) = 
	  List.map 
	    (function  r -> (r,VReg.newreg r) ) regs in
	let ret  = interp2 (new_vreg_env@vreg_env,val_env) rexp in
	List.iter (function (_,r) -> VReg.del r) new_vreg_env;
	ret
  with 
    Failure msg -> 
      (* print_string " -- stack trace : "; simple_print_r2exp e;  *)
      failwith msg
