open Rtype
open RSyntax
open R2Syntax

type label = string

type kexp =
    KVal of rvalue
  | KVar of ident
  | KFApp of ident * region list
  | KLLambda of label * (ident list * (ident * region) option) * kexp * region
(*  | KRegLLambda of label * region list * kexp * region *)
  | KApp of ident * ident list
  | KLet of kbind * kexp
  | KLetrec of krecbind list * kexp 
  | KIf  of ident * kexp * kexp 
  | KLetregion of region list * kexp 
and kbind = { kid:ident ; kbody:kexp }
and krecbind = 
    K_Regfun of label * region list * kexp * region
  | K_VBind  of label * kexp * region

let args_to_list =
  function 
      (idlst,Some(op,r)) -> idlst@[op]
    | (idlst,None)    -> idlst

let new_name = Syntax.new_name

(* K Normal Form $B$X$NJQ49(B *)
let rec knormal (rexp:r2exp) : kexp =
  match rexp with
    R2_Val(v) -> KVal(v)
  | R2_Var(v) -> KVar(v)
  | R2_FApp(rf,reglst) ->
      let v = new_name () in
      KLet({kid=v ; kbody=knormal rf},
	   KFApp(v, reglst))
  | R2_Lambda((arg,tail),body,reg) ->
      let label = new_name () in
      KLLambda(label,
	       (arg,tail),
	       knormal body,
	       reg)
  | R2_Reglambda(aregs,body,reg) ->
      failwith "KNormal.knormal : should not happen R2_Reglambda"
      (* let label = new_name () in
      KRegLLambda(label,
		  aregs,
		  knormal body,
		  reg)*)
  | R2_App(f,args) ->
      let fbind = {kid=new_name (); kbody= knormal f} in
      let argbind = List.map (function x -> {kid=new_name (); kbody= knormal x}) args in
      let rec binds_to_let binds = 
	begin
	  match binds with
	    hd::tl ->   KLet(hd,binds_to_let tl)
	  | [] ->  
	      KApp(fbind.kid,
		   List.map (function x -> x.kid) argbind)
	end
      in binds_to_let (fbind::argbind)
  | R2_Let(binds,body) ->
      let rec binds_to_let binds = 
	begin
	  match binds with
	    (r2id,r2body)::tl ->   KLet({kid=r2id; kbody=knormal r2body},
					binds_to_let tl)
	  | [] ->  
	      knormal body
	end
      in binds_to_let binds
  | R2_Letrec(binds,body) ->
      KLetrec(List.map 
		(function 
		    RB_Regfun(id,regs,r2body,reg) -> 
		      K_Regfun(id,regs,knormal r2body,reg)
		  | RB_VBind(id,r2body,reg) ->
		      K_VBind(id,knormal r2body, reg)) binds,
	      knormal body)
  | R2_If(e1,e2,e3) ->
      let newname = new_name () in
      KLet({kid=newname;  kbody=knormal e1},
	   KIf(newname, knormal e2,knormal e3))
  | R2_Letregion(regs, body) ->
      KLetregion(regs,knormal body)
