(** 饤֥ *)

type scmnum = 
    S_Int of int
  | S_Real of float

let string_of_scmnum n = 
  match n with
    S_Int(i) -> string_of_int i
  | S_Real(r) -> string_of_float r
    
let scm_zero = S_Int(0)
let scm_one = S_Int(1)

let num_binop fi ff a b =
  match (a,b) with
    (S_Int i1,S_Int i2) ->
      S_Int(fi i1 i2)
  | (S_Real i1,S_Int i2) ->
      S_Real(ff i1 (float_of_int i2))
  | (S_Int i1,S_Real i2) ->
      S_Real(ff (float_of_int i1) i2)
  | (S_Real i1,S_Real i2) ->
      S_Real(ff i1 i2)

let to_float n =
  match n with
    S_Int(a) -> float_of_int a
  | S_Real(a) -> a

let scm_plus = num_binop (+) (+.) 
let scm_minus = num_binop (-) (-.) 
let scm_mul = num_binop ( * ) ( *. ) 
let scm_div a b = S_Real((to_float a) /. (to_float b))
  

let num_binop_bool fi ff a b =
  match (a,b) with
    (S_Int i1,S_Int i2) ->
      fi i1 i2
  | (S_Real i1,S_Int i2) ->
      ff i1 (float_of_int i2)
  | (S_Int i1,S_Real i2) ->
      ff (float_of_int i1) i2
  | (S_Real i1,S_Real i2) ->
      ff i1 i2
let scm_lt = num_binop_bool (<) (<) 
let scm_le = num_binop_bool (<=) (<=) 
let scm_gt = num_binop_bool (>) (>) 
let scm_ge = num_binop_bool (>=) (>=) 
let scm_eq = num_binop_bool (=) (=) 


let is_real v = match v with S_Int _ -> false | S_Real _ -> true
let is_integer v = match v with S_Int _ -> true | S_Real _ -> false

let real_fun f v =
  match v with
    S_Int v ->  S_Real(f (float_of_int v))
  | S_Real v ->  S_Real(f v)

let scm_exp = real_fun exp
let scm_log = real_fun log
let scm_sin = real_fun sin
let scm_cos = real_fun cos
let scm_tan = real_fun tan
let scm_asin = real_fun asin
let scm_acos = real_fun acos
let scm_atan = real_fun atan
let scm_sqrt = real_fun sqrt

let real_to_int_fun f v =
  match v with
    S_Int v ->  S_Int(int_of_float(f (float_of_int v)))
  | S_Real v ->  S_Int(int_of_float (f v))

let scm_floor = real_to_int_fun floor
let scm_ceiling = real_to_int_fun ceil
let scm_truncate v = 
  match v with
    S_Int v ->  S_Int(truncate (float_of_int v))
  | S_Real v ->  S_Int(truncate v)
  
let scm_abs n = 
  match n with
    S_Int v ->  S_Int(abs v)
  | S_Real v ->  S_Real(abs_float v)
