%{
(* Ūѿؿʤɤ *)
open PreSyntax
open Library

exception Parse_error

let debug str = () (* print_endline ("paser : "^ str) *)

type tmp_cond =
    C_Clause of cond_clause
  | C_Else of expression list

%}

/* ȡ */

/* syntactic keyword */
%token ELSE RIGHTARROW DEFINE UNQUOTE UNQUOTE_SPLICING
/* syntactic_keyword - expressioin keyword */
%token QUOTE LAMBDA IF SET BEGIN COND AND OR CASE LET LETSTAR LETREC DO DELAY QUASIQUOTE LET_SYNTAX LETREC_SYNTAX DEFINE_SYNTAX SYNTAX_RULES

/* token */
%token LPAREN RPAREN SHARP_LPAREN S_QUOTE S_UNQUOTE S_COMMA S_COMMA_ATMARK DOT 
/* end_of_file */
%token EOF

%token <Library.scmnum> NUMBER
%token <bool>   BOOLEAN
%token <string> VARIABLE
%token <char>   CHARACTER
%token <string> STRING

/* ̤ͥassociativity(㤤⤤) */

/* ϵ */
%type <PreSyntax.program> program
%type <PreSyntax.definition> definition
%type <PreSyntax.binding_spec> binding_spec
%type <PreSyntax.expression> expression
%type <tmp_cond> cond_clause
%start program

%%
expression:
| VARIABLE
    { `Variable($1) }
| literal
    { `Literal($1) }
| procedure_call
    { $1 }
| lambda_expression
    { $1 }
| conditional
    { $1 }
| assignment
    { $1 }
| derived_expression
    { `Derived_expression $1 }
| macro_block
    { failwith "not implemented macro_block" }
;

literal:
| quotation 
    { `Datum $1 }
| self_evaluating
    { `Self_evaluating $1 }
;

self_evaluating:
|  BOOLEAN
    { `Boolean($1) }
|  NUMBER
    { `Number($1) }
|  CHARACTER
    { `Character($1) }
|  STRING 
    { `String($1) }
;

quotation:
| S_QUOTE datum
    { $2 }
| LPAREN QUOTE datum RPAREN
    { $3 }
;

procedure_call:
| LPAREN operator operand_list RPAREN
    { `Procedure_call ($2,$3) }
;

operator:
|  expression
    { $1 }
;

operand:
|  expression
    { $1 }
;

operand_list:
| operand_list operand 
    { $1@[$2] }
|  { [] }
;

lambda_expression:
|  LPAREN LAMBDA formals body RPAREN
    { `Lambda_expression($3,$4) }
;

formals:
| LPAREN variable_list RPAREN
    { ($2,None) }
| VARIABLE
    { ([],Some($1)) }
| LPAREN VARIABLE variable_list DOT VARIABLE RPAREN
    { ($2::$3, Some($5)) }
;

variable_list:
| VARIABLE variable_list
    { $1::$2 }
| 
  { [] }
;

body:
| command_or_definition_list
    { (* definitionθ˥ޥɤ³Ƥ뤫ɤΤ *)
      let rec is_coms cs =
	match cs with
	  `Command e::tl ->
	    is_coms tl
	| [] -> true
	| _ -> false
      in 
      let rec rem_begin dclist = 
	List.fold_right (@)
	  (List.map 
	     (function `Command e -> [`Command e]
	       | `Define d -> [`Define d]
	       | `P_Begin lst -> rem_begin lst)
	     dclist)
	  []
      in 
      let rec get_defs ds =
	match ds with
	  `Command e::tl ->
	    if is_coms ds
	    then
	      ([],List.map (function `Command e -> e | _ -> failwith "parser") ds)
	    else 
	      failwith "parser"
	| `Define d::tl -> 
	    let (ds2,cs) = get_defs tl in
	    (d::ds2,cs)
	| [] ->
	    ([],[])
	| _ -> failwith "parser"
      in 
      get_defs (rem_begin $1)

    }
;

sequence:
| expression command_list 
    { $1::$2 }
;

command:
| expression
    { $1 }
;
command_list:
| command command_list 
    { $1::$2 }
| 
    { [] }

conditional:
| LPAREN IF test consequent alternate RPAREN
    { `Conditional_if($3,$4,$5) }
;
test:
| expression
    { debug "test"; $1 }
;
test_list:
| test test_list
    { $1::$2 }
| 
    { [] }

consequent:
| expression
    { $1 }
;
alternate:
| expression
    { Some $1 }
| 
    { None }
;
assignment:
| LPAREN SET VARIABLE expression RPAREN
    { debug "set!"; `Assignment($3,$4) }
;
derived_expression:
| LPAREN COND cond_clause_list RPAREN
    { let rec iter c =
      match c with
	C_Clause a::[] -> ([a],[]) 
      | C_Else e ::[] -> ([],e) 
      |	C_Clause a::tl -> let (ctl,el) = iter tl in (a::ctl,el)
      | _ -> failwith "cond parsing"
    in 
    `Cond (iter $3) }
| LPAREN CASE expression case_clause case_clause_list RPAREN
    { failwith "not implemented case" }
| LPAREN AND test_list RPAREN
    { `And($3) }
| LPAREN OR test_list RPAREN
    { `Or($3) }
| LPAREN LET LPAREN binding_spec_list RPAREN body RPAREN
    { `Let($4,$6) }
| LPAREN LET VARIABLE LPAREN binding_spec_list RPAREN body RPAREN 
    { failwith "not implemented named_let" }
| LPAREN LETSTAR LPAREN binding_spec_list RPAREN body RPAREN
    { `LetStar($4,$6) }
| LPAREN LETREC LPAREN binding_spec_list RPAREN body RPAREN
    { `Letrec($4,$6) }
| LPAREN BEGIN sequence RPAREN
    { `Begin($3) }
| LPAREN DO LPAREN iteration_spec_list RPAREN LPAREN test do_result RPAREN command_list RPAREN
    { `Do($4,($7,$8),$10) }
| LPAREN DELAY expression RPAREN
    { `Delay }
;

cond_clause:
| LPAREN test sequence RPAREN
    { debug "cond_clause"; C_Clause($2,$3) }
| LPAREN test RPAREN
    { debug "cond_clause"; C_Clause($2,[]) }
| LPAREN test RIGHTARROW recipient RPAREN
    { debug "cond_clause"; C_Clause($2,[$4]) }
| LPAREN ELSE sequence RPAREN
    { debug "cond_clause"; C_Else($3) }
;
cond_clause_list:
| cond_clause cond_clause_list
    { $1::$2 }
| 
    { [] }
;
recipient:
| expression
    { $1 }
;
case_clause:
| LPAREN LPAREN datum_list RPAREN sequence RPAREN
    { failwith "not implemented case clause" }
;
case_clause_list:
| 
    { [] }
| case_clause case_clause_list
    { $1::$2 }
;
binding_spec:
| LPAREN VARIABLE expression RPAREN
    { ($2,$3) }
;
binding_spec_list:
| binding_spec binding_spec_list
    { $1::$2 }
| 
    { [] }
;
iteration_spec:
| LPAREN VARIABLE init step RPAREN
    { ($2,$3,$4) }
| LPAREN VARIABLE init RPAREN
    { ($2,$3,`Do_Nothing) }
;
iteration_spec_list:
| iteration_spec iteration_spec_list
    { $1::$2 }
| 
    { [] }

init:
| expression
    { $1 }
;
step:
| expression
    { $1 }
;
do_result:
| sequence
    { $1 }
| 
    { [] }
;
syntax_spec:
| LPAREN keyword transformer_spec RPAREN
    { failwith "not implemented syntax spec" }
;
syntax_spec_list:
| syntax_spec_list syntax_spec 
    { failwith "not implemented syntax spec list" }
| 
    { [] }

program:
| command_or_definition_list EOF
    { $1 }
command_or_definition:
| command
    { `Command $1 }
| definition
    { `Define $1 }
| syntax_definition
    { failwith "not implemented commmand or definition" }
;
definition:
| LPAREN DEFINE VARIABLE expression RPAREN
    { `DefineV($3,$4) }
| LPAREN DEFINE LPAREN VARIABLE def_formals RPAREN body RPAREN
    { `DefineF($4,$5,$7) }
;
command_or_definition_list:
| command_or_definition command_or_definition_list
    { $1::$2 }
| 
    { [] }

def_formals:
| variable_list
    { ($1,None) }
| variable_list DOT VARIABLE
    { ($1,Some($3)) }
;
syntax_definition:
| LPAREN DEFINE_SYNTAX keyword transformer_spec RPAREN
    { failwith "not implemented syntax definition" }
;    
keyword:
| VARIABLE
    { failwith "not implemented keyword" }
;

datum:
| simple_datum
    { $1 }
| compound_datum
    { $1 }
;
simple_datum:
| BOOLEAN 
    { `Boolean $1 }
| NUMBER
    { `Number $1 }
| CHARACTER
    { `Character $1 }
| STRING
    { `String $1 }
| symbol
    { `Symbol $1 }
;
symbol:
| VARIABLE
    { $1 }
;
compound_datum:
| list
    { $1 }
| vector
    { $1 }
;
list:
| LPAREN datum_list RPAREN
    { `List $2 }
| LPAREN datum datum_list DOT datum RPAREN
    { failwith "not implemented list " }
| abbreviation
    { failwith "not implemented list" }
;
abbreviation:
| QUOTE datum
    { failwith "not implemented quote" }
| UNQUOTE datum
    { failwith "not implemented unquote" }
| S_COMMA datum
    { failwith "not implemented scomma" }
| S_COMMA_ATMARK datum
    { failwith "not implemented scomma atmark" }
;
vector:
| SHARP_LPAREN datum_list RPAREN
    { `Vector $2 }
;
datum_list:
| datum datum_list
    { $1::$2 }
| 
    { [] }

macro_block:
| LPAREN LET_SYNTAX LPAREN syntax_spec_list RPAREN body RPAREN
    { failwith "not implemented macro block" }
| LPAREN LETREC_SYNTAX LPAREN syntax_spec_list RPAREN body RPAREN
    { failwith "not implemented macro block" }
;
transformer_spec:
|  LPAREN SYNTAX_RULES LPAREN variable_list RPAREN RPAREN
    { failwith "not implemented transformer spec" }
;
