open Library

type variable = string

type datum = [
    `Boolean   of bool
  | `Number    of scmnum
  | `Character of char
  | `String    of string
  | `Symbol    of string
  | `List      of datum list
  | `Vector    of datum list
  | `Abbrev    of abbrev
  ]
and abbrev = [
    `Quote    of datum
  | `Unquote  of datum
  | `Comma    of datum
  | `Comna_At of datum
]

type expression = [
    `Variable of variable
  | `Literal of literal
  | `Procedure_call of expression * expression list
  | `Lambda_expression of (variable list * variable option) * body
  | `Conditional_if of expression * expression * expression option
  | `Assignment of variable * expression
  | `Derived_expression of derived_expression
  | `Macro_use 
  | `Macro_block 
  | `Do_Nothing
]
and literal = [
    `Datum of datum
  | `Self_evaluating of self_evaluating
]
and self_evaluating = [
    `Boolean of bool
  | `Number of scmnum
  | `Character of char
  | `String of string
]
and body = definition list * expression list
and derived_expression = [
    `Cond of cond_clause list * expression list
  | `Case of expression * case_clause list * expression list
  | `And of expression list
  | `Or of expression list
  | `Let of binding_spec list * body
  | `NamedLet of variable * binding_spec list * body
  | `LetStar of binding_spec list * body
  | `Letrec of binding_spec list * body
  | `Begin of expression list
  | `Do of iteration_spec list * (test * expression list) * expression list
  | `Delay
  | `Quasiquotation
]
and definition = [
    `DefineV of variable * expression
  | `DefineF of variable * (variable list * variable option) * body
  | `D_Begin of definition list
  ]
and cond_clause = expression * expression list
and case_clause = unit
and binding_spec = variable * expression
and test = expression
and iteration_spec = variable * expression * expression 

type command_or_definition = [
    `Command of expression
  | `Define of definition
  | `P_Begin of command_or_definition list
]
type program = command_or_definition list


