(** $B7?$+$i%j!<%8%g%s$N>pJs$N$_$r<h$j=P$7$?9=J8!#(B*)

open Rtype
open RSyntax

type r2exp =
    R2_Val of rvalue
  | R2_Var of ident
  | R2_FApp of r2exp * region list
  | R2_Lambda of (ident list * (ident * region) option) * r2exp * region
  | R2_Reglambda of region list * r2exp * region 
  | R2_App of r2exp * r2exp list
  | R2_Let of r2bind list * r2exp
  | R2_Letrec of r2recbind list * r2exp 
  | R2_If  of r2exp * r2exp * r2exp 
  | R2_Letregion of region list * r2exp 
and r2bind = ident * r2exp 
and r2recbind = 
    RB_Regfun of ident * region list * r2exp * region 
  | RB_VBind of ident * r2exp * region 
	
let rec rexp_to_r2exp rexp =
  match rexp with
    RVal(v,_,_) -> R2_Val v
  | RVar(v,_,_) -> R2_Var v
  | RFVar(v,rlst,_,_,_) -> R2_FApp(R2_Var(v),rlst)
  | RLambda((args,tail),rexp,reg,rt,_) -> 
      let newtail = 
	begin 
	  match tail with
	    Some tvar -> 
	      let tailregion = 
		begin 
		  match rt.func with
		    Some((_,ttl),_,_,_) ->
		      begin 
			match ttl.pair with
			  Some(_,_,reg) -> reg
			| _ -> failwith "rexp to r2exp"
		      end
		  | _ -> failwith "rexp to r2exp"
		end 
	      in Some(tvar,tailregion)
	  | None -> None
	end 
      in 
      R2_Lambda((args,newtail), rexp_to_r2exp rexp, reg)
  | RApp(f,args,_,_) -> R2_App(rexp_to_r2exp f,
			       List.map rexp_to_r2exp args)
  | RLet(binds,exp,_,_) -> R2_Let(List.map rbind_to_r2bind binds,
				  rexp_to_r2exp exp)
  | RLetrec(binds,exp,_,_) -> R2_Letrec(List.map rbind_to_r2recbind binds,
					rexp_to_r2exp exp)
  | RIf(e1,e2,e3,_,_) -> R2_If(rexp_to_r2exp e1,
			       rexp_to_r2exp e2,
			       rexp_to_r2exp e3)
  | RLetregion(regs,exp,_,_) -> R2_Letregion(Rtype.regs_in_effect regs,
					     rexp_to_r2exp exp)
and rbind_to_r2bind bind = 
  match bind.rt with
    Forall(regs,_,_,_,reg) ->
      (bind.rid, R2_Reglambda(regs, rexp_to_r2exp bind.rbody, reg))
  | Normal(_) ->
      (bind.rid,rexp_to_r2exp bind.rbody)
and rbind_to_r2recbind bind = 
  match bind.rt with
    Forall(regs,_,_,_,reg) ->
      RB_Regfun(bind.rid, regs, 
		rexp_to_r2exp bind.rbody,
		reg)
  | Normal(t) ->
      begin
	match (Rtype.topLevel t) with
	  [reg] ->
	    RB_VBind(bind.rid,
		     rexp_to_r2exp bind.rbody, reg)
	| _ -> failwith ("invalid letrec bind type.  Is your inference rule is correct?" 
			 ^ (string_of_rtype_scheme bind.rt))
      end 

open Format
  
let simple_print_r2exp (e:r2exp) =
  let rec pr_rexp ppf rexp = 
    match rexp with 
      R2_Val(v) -> fprintf ppf "%s" 
	  (string_of_rvalue v)
    | R2_Var(v) ->  fprintf ppf "%s" v
    | R2_FApp(v,regs) ->  
	fprintf ppf "%a[%s]" 
	  pr_rexp v (List.fold_right (^) (List.map string_of_reg regs) "") 
    | R2_Lambda((args,Some (t,treg)),body,reg) ->
	let sargs = (List.fold_right (fun x y -> x^" "^y)  args "")
	    ^ "." ^ t ^ "_at_" ^ (string_of_reg treg )
	in
	fprintf ppf "@[<hov 3>(lambda@ %s@,%a)@ at@ %s@]@," 
	  ("("^sargs^")") pr_rexp body (string_of_reg reg)
    | R2_Lambda((args,None),body,reg) ->
	let sargs = List.fold_right (fun x y -> x^" "^y)  args "" in
	fprintf ppf "@[<hov 3>(lambda@ %s@,%a)@ at@ %s@]@," 
	  ("("^sargs^")") pr_rexp body (string_of_reg reg)
    | R2_Reglambda(regs,body,reg) ->
	let sargs = List.fold_right (fun x y -> x^" "^y)  (List.map string_of_reg regs) "" in
	fprintf ppf "@[<hov 3>(reglambda@ %s@,%a)@ at@ %s@]@," 
	  ("("^sargs^")") pr_rexp body (string_of_reg reg)
    | R2_App(f,args) ->
	fprintf ppf "(@[<hov 0>%a@ %a@])@," 
	  pr_rexp f pr_lst args 
    | R2_Let(rbindlst,body) ->
	fprintf ppf "@[<hov 3>(let@ (@[<h 0>%a)@]@,%a)@,@]" 
	  pr_rblst rbindlst pr_rexp body 
    | R2_Letrec(rbindlst,body) ->
	fprintf ppf "@[<hov 3>(letrec@ (@[<h 0>%a)@]@,%a)@,@]" 
	  pr_rbreclst rbindlst pr_rexp body 
    | R2_If(e1,e2,e3) ->
	fprintf ppf "(if@[<hov 0>@ %a@ @,%a@ @,%a@])" 
	  pr_rexp e1 pr_rexp e2 pr_rexp e3 
    | R2_Letregion(regs,body) ->
	let sargs = List.fold_right (fun x y -> x^" "^y)  (List.map string_of_reg regs) "" in
	fprintf ppf "@[<hov 3>(letregion@ (%s)@ @,%a)@]" 
	  sargs pr_rexp body 
  and pr_lst ppf =
    function 
	hd::tl -> fprintf ppf  "%a@ %a"  pr_rexp hd pr_lst tl
      | [] -> fprintf ppf "%s" ""
  and pr_rbind ppf (id,body) = 
    fprintf ppf "@[<hov 1>(%s@,@ %a)@]"
      id pr_rexp body
  and pr_rblst ppf =
    function 
	hd::tl -> fprintf ppf "@[<hov 0>%a@;%a@]"  pr_rbind hd pr_rblst tl
      | [] -> fprintf ppf "%s" ""
  and pr_rrecbind ppf b = 
    match b with
      RB_Regfun(id,regs,body,r) ->
	let sargs = List.fold_right (fun x y -> x^" "^y)  (List.map string_of_reg regs) "" in
	fprintf ppf "@[<hov 1>(%s[%s]@ =@ %a)@ at@ %s@ @]" 
	  id sargs pr_rexp body (string_of_reg r)
    | RB_VBind(id,body,reg) -> 
	fprintf ppf "@[<hov 1>(%s@ =@ %a)@ at@ %s@ @]" 
	  id pr_rexp body (string_of_reg reg)
  and pr_rbreclst ppf =
    function 
	hd::tl -> fprintf ppf "@[<hov 0>%a@;%a@]"  pr_rrecbind hd pr_rbreclst tl
      | [] -> fprintf ppf "%s" ""
  in 
  printf "%a@." pr_rexp e; print_string "\n"
    
 
let allocates_global_regions (re:r2exp) : r2exp = 
  let rec global_regions re = 
    match re with
      R2_Val(RNumber(_,reg)) -> [reg] 
    | R2_Val(RCharacter(_,reg)) -> [reg] 
    | R2_Val(RString(_,reg)) -> [reg] 
    | R2_Val(RSymbol(_,reg)) -> [reg] 
    | R2_Val(_) -> [] 
    | R2_Var(_) -> []
    | R2_FApp(rf,regs) -> (global_regions rf)@regs
    | R2_Lambda((_,_),body,reg) -> reg::(global_regions body)
    | R2_Reglambda(reglist,body,reg) -> reg::(Mylist.subtract_list (global_regions body) reglist)
    | R2_App(f,args) -> 
	Mylist.uniq (List.fold_right (@) (List.map global_regions (f::args)) [])
    | R2_Let(r2blist,body) -> (global_r2binds r2blist)@(global_regions body)
    | R2_Letrec(r2blist,body) -> (global_r2recbinds r2blist)@(global_regions body)
    | R2_If(e1,e2,e3) -> (global_regions e1)@(global_regions e2)@(global_regions e3)
    | R2_Letregion(regs,body) -> (Mylist.subtract_list (global_regions body) regs)
  and global_r2binds rblst = 
    List.fold_right (@) 
      (List.map (function (id,body) -> global_regions body) rblst) []
  and global_r2recbinds rblst = 
    List.fold_right (@) 
      (List.map 
	 (function 
	     RB_Regfun(id,regs,body,r) ->
	       r::(Mylist.subtract_list (global_regions body) regs)
	   | RB_VBind(id,body,reg) -> global_regions body) rblst)
      []
      
  in
  R2_Letregion (Mylist.uniq (global_regions re),re)
