(* $Id: rSyntax.ml,v 1.22 2003/10/11 15:48:50 ganat Exp $ *)

open TSyntax
open Rtype
open Library 

type ident = string

type rvalue =
    RNumber of scmnum * region
  | RBoolean of bool 
  | RCharacter of char * region
  | RString of string * region
  | RSymbol of string * region
  | RNull 
  | RUndef

(* ꡼Ĥ󥿥å *) 
type rexp =
    RVal of rvalue * rtype * effect
  | RVar of ident * rtype * effect
  | RFVar of ident * region list * effvar list * rtype * effect 
  | RLambda of (ident list * ident option) * rexp * region * rtype * effect
  | RApp of rexp * rexp list * rtype * effect 
  | RLet of rbind list * rexp * rtype * effect 
  | RLetrec of rbind list * rexp * rtype * effect
  | RIf  of rexp * rexp * rexp * rtype * effect
(*  | RPair of rexp * rexp * region * rtype * effect *)
  | RLetregion of effect * rexp * rtype * effect 
and rbind = { rid:ident ; rt:rtype_scheme ; rbody:rexp }


(* let size (re:rexp) : effect = *)

      
let getEffect (re:rexp) : effect =
  match re with
    RVal(_,_,e) ->e 
  | RVar(_,_,e) ->e 
  | RFVar(_,_,_,_,e) ->e 
  | RLambda(_,_,_,_,e) ->e 
  | RApp(_,_,_,e) ->e  
  | RLet(_,_,_,e) ->e  
  | RLetrec(_,_,_,e) ->e  
  | RIf(_,_,_,_,e) ->e  
(*  | RPair(_,_,_,_,e) ->e  *)
  | RLetregion(_,_,_,e) -> e
	
let getRtype (re:rexp) : rtype =
  match re with
    RVal(_,r,_) -> r
  | RVar(_,r,_) -> r
  | RFVar(_,_,_,r,_) -> r
  | RLambda(_,_,_,r,_) -> r 
  | RApp(_,_,r,_) -> r
  | RLet(_,_,r,_) -> r
  | RLetrec(_,_,r,_) -> r
  | RIf(_,_,_,r,_) -> r
(*  | RPair(_,_,_,r,_) -> r *)
  | RLetregion(_,_,r,_) -> r


(* ɽؿ *)
let cat a b = a ^ " " ^b


(* ȥեȤɽ *)
let string_of_t_and_e (t:rtype) (e:effect) =
  "[" ^ (string_of_rtype t) ^ "," ^ (string_of_effect e) ^ "]"

(* rvalue ʸؤѴ *)
let string_of_rvalue rval =
  match rval with
    RNumber(n,reg) -> (string_of_scmnum n) ^ " at " ^ (string_of_reg reg)
  | RBoolean(b) -> if b then "#t" else "#f"
  | RCharacter(c,reg) -> let s = "a" in s.[0] <- c ; "#\\" ^ s ^ " at " ^  (string_of_reg reg)
  | RString(str,reg) -> str ^ " at " ^  (string_of_reg reg)
  | RSymbol(str,reg) -> "'" ^ str ^ " at " ^  (string_of_reg reg)
  | RNull -> "'()"
  | RUndef -> "*undef*"


(* ɽ , flg  trueλϷȥեȤɽ  *)
let rec string_of_rexp (flg:bool) =
  function
      RVal(v,rt,ef) -> 
	(string_of_rvalue v) ^ (if flg then ":" ^ (string_of_t_and_e rt ef) else "")
    | RVar(v,rt,ef) -> v ^ (if flg then ":" ^ (string_of_t_and_e rt ef) else "")
    | RFVar(v,rlst,elst,rt,ef) ->
	v ^ "[" ^ (List.fold_right cat (List.map string_of_reg rlst) "") 
	^ ";" ^(List.fold_right cat elst "") ^"]" 
	^ (if flg then ":" ^ (string_of_t_and_e rt ef) else "")
    | RLambda((lblst,Some tail),re,reg,rt,ef) ->
	"(lambda (" ^ (List.fold_right cat lblst "") ^ "." ^ tail
	^ ")" ^(string_of_rexp flg re) ^ ")," ^  (string_of_reg reg) ^ (if flg then ":" ^ (string_of_t_and_e rt ef) else "")
    | RLambda((lblst,None),re,reg,rt,ef) ->
	"(lambda (" ^ (List.fold_right cat lblst "")
	^ ")" ^(string_of_rexp flg re) ^ ")," ^  (string_of_reg reg) ^ (if flg then ":" ^ (string_of_t_and_e rt ef) else "")
    | RApp(rf,relst,rt,ef) ->
	(string_of_rexp flg rf) 
	^ (List.fold_right (^) (List.map (string_of_rexp flg) relst) "")
	^ (if flg then ":" ^ (string_of_t_and_e rt ef) else "")
    | RLet(rblst,re,rt,ef) ->
	"(let (" ^ (List.fold_right cat (List.map (string_of_rbind flg) rblst) "")
	^ ")" ^ (string_of_rexp flg re) ^ ")" ^ (if flg then ":" ^ (string_of_t_and_e rt ef) else "")
    | RLetrec(rblst,re,rt,ef) ->
	"(letrec (" ^ (List.fold_right cat (List.map (string_of_rbind flg) rblst) "")
	^ ")" ^ (string_of_rexp flg re) ^ ")" ^ (if flg then ":" ^ (string_of_t_and_e rt ef) else "")
    | RIf(re1,re2,re3,rt,ef) ->
	"(if " ^ (List.fold_right cat (List.map (string_of_rexp flg) [re1;re2;re3]) "")
	^ ")"  ^ (if flg then ":" ^ (string_of_t_and_e rt ef) else "")
(*    | RPair(re1,re2,reg,rt,ef) ->
	"(" ^ (string_of_rexp flg re1) ^ " . " ^ (string_of_rexp flg re2) ^ ")" ^ "," ^ reg 
	 ^ (if flg then ":" ^ (string_of_t_and_e rt ef) else "") *)
    | RLetregion(regs,re,rt,ef) ->
	"(letregion (" ^ (string_of_effect regs) ^ ")" ^
	(string_of_rexp flg re) ^ (if flg then ":" ^ (string_of_t_and_e rt ef) else "")

and string_of_rbind flg rb =
  "(" ^ rb.rid ^ (if flg then ":" ^ (string_of_rtype_scheme rb.rt)  else simpleString_of_RtypeScheme rb.rt)
  ^ " " ^ (string_of_rexp flg rb.rbody) ^ ")"
and simpleString_of_RtypeScheme rs =
  match rs with
    Forall(rlst,rpeff,eflst,rt,reg) -> ":\\" ^ (List.fold_right (^) (List.map string_of_reg rlst) "") ^ "," ^ (string_of_reg reg) ^ " "
  | Normal(_) -> ""
    


(* ------------------------------------------------- *)  
open Format
let print_rexp (e:rexp) =
  let rec pr_rexp ppf rexp = 
    match rexp with 
      RVal(v,rt,ef) -> fprintf ppf "%s:%s&%s" 
	  (string_of_rvalue v) (string_of_rtype rt) (string_of_effect ef)
    | RVar(v,rt,ef) ->  fprintf ppf "%s:%s&%s" v (string_of_rtype rt) 
	  (string_of_effect ef)
    | RFVar(v,regs,evs,rt,ef) ->  
	fprintf ppf "%s[%s,%s]:%s&%s" 
	  v (List.fold_right (^)  (List.map string_of_reg regs) "") (List.fold_right (^) evs "")
	  (string_of_rtype rt) (string_of_effect ef)
    | RLambda((args,tail),body,reg,rt,ef) ->
	let sargs = (List.fold_right (fun x y -> x^" "^y)  args "")
	    ^ (match tail with Some t -> " . " ^ t | None -> "")
	in
	fprintf ppf "@[<hov 3>(lambda@ %s@ %a)@ at@ %s@,:%s&%s@]" 
	  ("("^sargs^")") pr_rexp body  (string_of_reg reg) (string_of_rtype rt) 
	  (string_of_effect ef)
    | RApp(f,args,rt,ef) ->
	fprintf ppf "(@[<hov 0>%a@ %a@])@,:%s&%s" 
	  pr_rexp f pr_lst args (string_of_rtype rt)
	  (string_of_effect ef)
    | RLet(rbindlst,body,rt,ef) ->
	fprintf ppf "@[<hov 3>(let@ (@[<h 0>%a)@]@,%a)@,:%s&%s@]" 
	  pr_rblst rbindlst pr_rexp body (string_of_rtype rt)
	  (string_of_effect ef)
    | RLetrec(rbindlst,body,rt,ef) ->
	fprintf ppf "@[<hov 3>(letrec@ (@[<h 0>%a)@]@,%a)@,:%s&%s@]" 
	  pr_rblst rbindlst pr_rexp body (string_of_rtype rt)
	  (string_of_effect ef)
    | RIf(e1,e2,e3,rt,ef) ->
	fprintf ppf "@[<hov 0>(if@ %a@ %a@ %a)@,:%s&%s@]" 
	  pr_rexp e1 pr_rexp e2 pr_rexp e3 (string_of_rtype rt)
	  (string_of_effect ef)	  
(*    | RPair(e1,e2,reg,rt,ef) ->
	fprintf ppf "@[<hov 0>(%a@ .@ %a)@ at@ %s@,:%s&%s@]" 
	  pr_rexp e1 pr_rexp e2 reg (string_of_rtype rt)
	  (string_of_effect ef) *)
    | RLetregion(regs,body,rt,ef) ->	  
	fprintf ppf "@[<hov 3>(letregion@ (%s)@,%a)@,:%s&%s@]" 
	  (string_of_effect regs) pr_rexp body (string_of_rtype rt)
	  (string_of_effect ef)
  and pr_lst ppf =
    function 
	hd::tl -> fprintf ppf  "%a@ %a"  pr_rexp hd pr_lst tl
      | [] -> fprintf ppf "%s" ""
  and pr_rbind ppf b = 
    fprintf ppf "@[<hov 1>(%s:%s@,%a)@]" 
      b.rid (string_of_rtype_scheme b.rt) pr_rexp b.rbody 
  and pr_rblst ppf =
    function 
	hd::tl -> fprintf ppf "@[<hov 0>%a@,%a@]"  pr_rbind hd pr_rblst tl
      | [] -> fprintf ppf "%s" ""
  in 
  printf "%a@." pr_rexp e; print_string "\n"
    
  
let simple_print_rexp (e:rexp) =
  let rec pr_rexp ppf rexp = 
    match rexp with 
      RVal(v,rt,ef) -> fprintf ppf "%s" 
	  (string_of_rvalue v)
    | RVar(v,rt,ef) ->  fprintf ppf "%s" v
    | RFVar(v,regs,evs,rt,ef) ->  
	fprintf ppf "%s[%s,%s]" 
	  v (List.fold_right (^) (List.map string_of_reg regs) "") (List.fold_right (^) evs "")
    | RLambda((args,tail),body,reg,rt,ef) ->
	let sargs = (List.fold_right (fun x y -> x^" "^y)  args "") ^
	    (match tail with Some t -> "." ^ t | None -> "")
	in
	fprintf ppf "@[<hov 3>(lambda@ %s@,%a)@ at@ %s@]@," 
	  ("("^sargs^")") pr_rexp body (string_of_reg reg )
    | RApp(f,args,rt,ef) ->
	fprintf ppf "(@[<hov 0>%a@ %a@])@," 
	  pr_rexp f pr_lst args 
    | RLet(rbindlst,body,rt,ef) ->
	fprintf ppf "@[<hov 3>(let@ (@[<h 0>%a)@]@,%a)@,@]" 
	  pr_rblst rbindlst pr_rexp body 
    | RLetrec(rbindlst,body,rt,ef) ->
	fprintf ppf "@[<hov 3>(letrec@ (@[<h 0>%a)@]@,%a)@,@]" 
	  pr_rblst rbindlst pr_rexp body 
    | RIf(e1,e2,e3,rt,ef) ->
	fprintf ppf "(if@[<hov 0>@ %a@ @,%a@ @,%a@])" 
	  pr_rexp e1 pr_rexp e2 pr_rexp e3 
(*    | RPair(e1,e2,reg,rt,ef) ->
	fprintf ppf "@[<hov 0>(%a@ .@ %a)@ at@ %s@,@]" 
	  pr_rexp e1 pr_rexp e2 reg *)
    | RLetregion(regs,body,rt,ef) ->	  
	fprintf ppf "@[<hov 3>(letregion@ (%s)@ @,%a)@]" 
	  (string_of_effect regs) pr_rexp body 
  and pr_lst ppf =
    function 
	hd::tl -> fprintf ppf  "%a@ %a"  pr_rexp hd pr_lst tl
      | [] -> fprintf ppf "%s" ""
  and pr_rbind ppf b = 
    fprintf ppf "@[<hov 1>(%s:%s@,%a)@]" 
      b.rid (string_of_rtype_scheme b.rt) pr_rexp b.rbody 
  and pr_rblst ppf =
    function 
	hd::tl -> fprintf ppf "@[<hov 0>%a@;%a@]"  pr_rbind hd pr_rblst tl
      | [] -> fprintf ppf "%s" ""
  in 
  printf "%a@." pr_rexp e; print_string "\n"
    
  




let isRValue (re:rexp) : bool =
  match re with
    RVal(_,_,_) -> false
  | RVar(_,_,_) -> false
  | RFVar(_,_,_,_,_) -> false
  | RLambda(_,_,_,_,_) -> true 
  | RApp(_,_,_,_) ->  false
  | RLet(_,_,_,_) ->  false
  | RLetrec(_,_,_,_) -> false
  | RIf(_,_,_,_,_) -> false
(*  | RPair(_,_,_,_,_) -> false *)
  | RLetregion(_,_,_,_) ->  false

let substToRvalue (s:rsubst) (rv:rvalue) : rvalue =
  match rv with 
    RNumber(str,reg) -> RNumber(str,substToRegion s reg)
  | RBoolean(b) -> rv
  | RCharacter(c,reg) -> RCharacter(c,substToRegion s reg)
  | RString(c,reg) -> RString(c,substToRegion s reg)
  | RSymbol(c,reg) -> RSymbol(c,substToRegion s reg)
  | RNull -> rv
  | RUndef -> rv
  

let rec substToRexp (s:rsubst) (re:rexp) : rexp = 
  match re with
    RVal(i,rt,e) -> 
      RVal(substToRvalue s i,substToType s rt,substToEffect s e) 
  | RVar(v,rt,e) -> 
      RVar(v,substToType s rt,substToEffect s e)
  | RFVar(v,rs,es,rt,e) -> 
      RFVar(v,
	    List.map (substToRegion s) rs,
	    List.map 
	      (function ev -> 
		 if(List.mem_assoc ev (snd s))
		 then fst (List.assoc ev (snd s))
		 else ev)
	      es,
	    substToType s rt,substToEffect s e)
  | RLambda(args,body,r,rt,e) ->  
      RLambda(args,
	      substToRexp s body,
	      substToRegion s r,
	      substToType s rt,substToEffect s e)
  | RApp(f,args,rt,e) ->
      RApp(substToRexp s f,
	   List.map (substToRexp s) args,
	   substToType s rt,substToEffect s e)
  | RLet(bl,body,rt,e) -> 
      RLet(List.map (substToRbind s) bl,
	   substToRexp s body,
	   substToType s rt,substToEffect s e) 
  | RLetrec(bl,body,rt,e) -> 
      begin
	match (substToRexp s (RLet(bl,body,rt,e))) with
	  RLet(bl,b,rt,e) -> RLetrec(bl,b,rt,e) 
	| _ -> failwith "RSintax.substToRexp : Unknown exception"
      end
  | RIf(a,b,c,rt,e) ->
      RIf(substToRexp s a,
	  substToRexp s b,
	  substToRexp s c,
	  substToType s rt,substToEffect s e) 	
(*  | RPair(a,b,reg,rt,e) ->
      RPair(substToRexp s a,substToRexp s b,
	    substToRegion s reg,
	    substToType s rt,substToEffect s e) *)
  | RLetregion(regs,re,rt,e) -> 
      let forletreg s x = 
	match x with
	  Reg(r) -> Reg (substToRegion s r)
	| Eff(e) -> 
	    if List.mem_assoc e (snd s)
	    then Eff (fst (List.assoc  e (snd s)))
	    else Eff e
      in 
      RLetregion(List.map (forletreg s) regs,
		 substToRexp s re,
		 substToType s rt,
		 substToEffect s e)
and substToRbind (s:rsubst) (rb:rbind) : rbind = 
  {rid  = rb.rid;
   rt   = substToRScheme s rb.rt; 
   rbody= substToRexp s rb.rbody}
	
	
let insertLetRegion (renv:rtype_env) (re:rexp) : rexp =
  let reserve_regs = 
    let (regs,effs) = Mylist.append2 (fv (getRtype re)) (fv_of_Env renv) in
    (List.map (function x -> Reg x) regs)@(List.map (function x -> Eff x) effs) in 
  let letregs = 
    Mylist.uniq (Mylist.subtract_list  (getEffect re) reserve_regs) in
  if(List.length letregs > 0)
  then 
    begin
      RLetregion(letregs, re, getRtype re,
		 (Mylist.subtract_list  (getEffect re) letregs))
    end
  else 
    begin
      re
    end
      
      
(* PI㤦Τˤ롣 *)
(* new_rtold_rt region polymorphic ǤʤƤϤʤʤ *)
(* new_rtold_rt old_rt generalǤʤƤϤʤʤ *)
(* ޤ generalize Ƥѿʳ ƱǤʤƤϤʤʤ*)
let replaceRscheme (f:string)  (old_rt:rtype_scheme) (new_rt:rtype_scheme) (arg_re:rexp) : rexp =
  (* print_endline ("replace  : " ^ f); *)
  match (old_rt,new_rt) with
    (Forall(regs1, rpe1, effs1, rt1, reg1),
     Forall(regs2, rpe2, effs2, rt2, reg2)) -> 
       if ((List.length regs1 < List.length regs2)
	 || (List.length effs1 < List.length effs2))
       then failwith ("RSyntax.replaceRscheme : invalid argument. 1\n " 
		      ^ (string_of_rtype_scheme old_rt) ^ " and " 
		      ^ (string_of_rtype_scheme new_rt))
       else 
	 let pairappend = fun x y -> ((fst x)@(fst y),(snd x)@(snd y)) in
	 (* ƵŪбѿ򸫤Ƥ *)
	 let rec findRec (donelst:id list) (rt1:rtype) (rt2:rtype) 
	     : ((region * region) list * (effvar * effvar) list) =
	   if(List.mem rt1.label donelst)
	   then ([] ,[])
	   else 
	     List.fold_right pairappend
	       [begin
		 match (rt1.num,rt2.num) with
		   (Some(r1),Some(r2)) -> 
		     ([(r1,r2)],[])
		 | _ -> ([],[])
	       end; 
		begin
		  match (rt1.char,rt2.char) with
		    (Some(r1),Some(r2)) -> 
		     ([(r1,r2)],[])
		  | _ -> ([],[])
		end; 
		begin
		 match (rt1.string,rt2.string) with
		   (Some(r1),Some(r2)) -> 
		     ([(r1,r2)],[])
		 | _ -> ([],[])
		end; 
		begin
		 match (rt1.symbol,rt2.symbol) with
		   (Some(r1),Some(r2)) -> 
		     ([(r1,r2)],[])
		 | _ -> ([],[])
	       end; 
		begin
		 match (rt1.port,rt2.port) with
		   (Some(r1),Some(r2)) -> 
		     ([(r1,r2)],[])
		 | _ -> ([],[])
	       end; 
		begin
		  match (rt1.func,rt2.func) with
		    (Some((ra1,tail1),(ev1,es1),rr1,reg1),
		     Some((ra2,tail2),(ev2,es2),rr2,reg2)) ->
		       List.fold_right pairappend
			 [List.fold_right pairappend 
			    (List.map2 (findRec (rt1.label::donelst)) ra1 ra2) ([],[]);
			  (findRec (rt1.label::donelst) tail1 tail2);
			  (findRec (rt1.label::donelst) rr1 rr2);
			  ([(reg1,reg2)],[]);
			  ([],[(ev1,ev2)]);
			] ([],[])
		  | (None,None) -> ([],[])
		  | _ -> failwith "RSyntax.replaceRscheme : invalid argument. 2"
		end;
		begin
		  match (rt1.vector,rt2.vector) with
		    (Some(v1,reg1),
		     Some(v2,reg2)) ->
		       List.fold_right pairappend
			 [(findRec (rt1.label::donelst) v1 v2);
			  ([(reg1,reg2)],[]) ;
			]  ([],[])
		  | (None,None) -> ([],[])
		  | _ -> failwith "RSyntax.replaceRscheme : invalid argument. 3"
		end;
		begin
		  match (rt1.pair,rt2.pair) with
		    (Some(car1,cdr1,reg1),
		     Some(car2,cdr2,reg2)) ->
		       List.fold_right pairappend
			 [(findRec (rt1.label::donelst) car1 car2);
			  (findRec (rt1.label::donelst) cdr1 cdr2);
			  ([(reg1,reg2)],[]) ;
			]  ([],[])
		  | (None,None) -> ([],[])
		  | _ -> failwith "RSyntax.replaceRscheme : invalid argument. 3"
		end
	      ] ([],[]) 
	 in 
	 let replst = findRec [] rt2 rt1 in
	 (* regs2 effs2 ΤΤѿ *)
	 let livelst = (List.map 
			  (function (x:region) -> Mylist.assoc x (fst replst)) regs2,
			List.map 
			  (function x -> Mylist.assoc x (snd replst)) effs2)
	 in
	 
	 (* ΤѿбȤ trueˤʤ *)
	 let liveflag = (List.map 
			   (function x -> List.mem x (fst livelst)) regs1,	
			 List.map 
			   (function x -> List.mem x (snd livelst)) effs1) in
	 let rec renewRvar (re:rexp) : rexp =
	   match re with 
	     RVal(_,_,_) -> re
	   | RVar(_,_,_) -> re
	   | RFVar(id, regs, effs, rt, eff) ->
	       if( id = f ) 
	       then
		 let rec iter (vars:region list) (lives:bool list) : region list =
		   match (vars,lives) with
		     (var::vtl,b::ltl) -> 
		       if b then var::(iter vtl ltl) else (iter vtl ltl)
		   | ([],[]) -> []
		   | _ ->  failwith "RSyntax.replaceRscheme : invalid argument. 4"
		 in 
		 let rec iter2 (vars:effvar list) (lives:bool list) : effvar list =
		   match (vars,lives) with
		     (var::vtl,b::ltl) -> 
		       if b then var::(iter2 vtl ltl) else (iter2 vtl ltl)
		   | ([],[]) -> []
		   | _ ->  failwith "RSyntax.replaceRscheme : invalid argument. 4"
		 in 
		 let ret = RFVar(id,
				 iter regs (fst liveflag),
				 iter2 effs (snd liveflag),
				 rt, eff)
		 in 
		 ret
	       else re
	   | RLambda(args,re,reg,rt,eff) -> 
	       RLambda(args,renewRvar re,reg,rt,eff) 
	   | RApp(re,relst,rt,eff) ->
	       RApp(renewRvar re,List.map renewRvar relst,rt,eff) 
	   | RLet(rblst,re,rt,eff) ->
	       RLet(List.map (function x-> {rid=x.rid;  rt=x.rt;
					    rbody=renewRvar x.rbody}) rblst,
		    renewRvar re,rt,eff)
	   | RLetrec(rblst,re,rt,eff) ->
	       RLetrec(List.map (function x-> {rid=x.rid;  rt=x.rt;
					       rbody=renewRvar x.rbody}) rblst,
		       renewRvar re,rt,eff)
	   | RIf(re1, re2, re3, rt, eff) ->
	       RIf(renewRvar re1,
		   renewRvar re2,
		   renewRvar re3, rt, eff) 
	   | RLetregion(regs,re,rt,eff) ->
	       RLetregion(regs,renewRvar re,rt,eff) 
	 in 
	 let r = renewRvar arg_re in
	 r
	 
  | (Normal(_),Normal(_)) -> arg_re
  | _ -> failwith "RSyntax.replaceRscheme : invalid argument. 5"
	
	

let rec fv_of_rexp (re:rexp) =
  let (lreg,leff) =
    match re with
      RVal(_,t,_) -> fv t
    | RVar(_,t,_) -> fv t
    | RFVar(_,regs,effs,t,_) -> Mylist.append2 (regs,effs) (fv t)
    | RLambda(_,body,r,t,e) -> 
	Mylist.append2 (Mylist.append2 (fv_of_rexp body) ([r],[])) (fv t)
    | RApp(f,args,t,e) ->
	List.fold_right Mylist.append2 (List.map fv_of_rexp (f::args)) ([],[])
    | RLet(bnds,body,t,e) ->
	List.fold_right Mylist.append2 
	  (List.map (function x -> fv_of_rexp x.rbody) bnds) (fv_of_rexp body)
    | RLetrec(bnds,body,t,e) ->
	List.fold_right Mylist.append2 
	  (List.map (function x -> fv_of_rexp x.rbody) bnds) (fv_of_rexp body)
    | RIf(e1,e2,e3,t,e) ->
	List.fold_right Mylist.append2 
	  (List.map  fv_of_rexp  [e1;e2;e3]) ([],[])
    | RLetregion(letregs,body,t,e) ->
	fv_of_rexp body 
  in 
  (Mylist.uniq lreg,Mylist.uniq leff)

	
