{
(* إå *)
(* Ūѿؿʤɤ *)
 open Read_parser
 open Library
 let line = ref 1
    
}
 
(* ɽά *)
let space = [' ' '\t']
let digit = ['0'-'9']
    
let delimiter = space | '(' | ')' | '"' | ';'
let whitespace = space 
let comment = ';' [^ '\n']*'\n'
let atmosphere = whitespace 
let intertoken_space = atmosphere* 
let intertoken_space_nl = comment | '\n'
    
let letter = ['a'-'z' 'A'-'Z']
let special_initial = '!'| '$' | '%' | '&' | '*' | '/' | ':' | '<' | '=' | '>' | '?' | '^' | '_' | '~'
let initial = letter | special_initial
let special_subsequent = '+' | '-' | '.' |'@'
let subsequent = initial | digit | special_subsequent
let peculiar_identifier = '+' | '-' | '*' | '/'
let identifier = initial subsequent* | peculiar_identifier
    
let expression_keyword = 
  "quote" | "lambda" | "if" | "set!" | "begin" | "cond" | "and" | "or" | "case" | 
  "let" | "let*" | "letrec" | "do" | "delay" | "quasiquote"
let syntactic_keyword = 
  expression_keyword | "else" | "=>" | "define" | "unquote" | "unquote-splicing"
    
let boolean = "#t" | "#f"
let character_name = "space" | "newline"
let character = "#\\" _ | "#\\" character_name
let string_element = [^ '"' '\\' ] | "\\\"" | "\\\\"
let string = '"' string_element* '"' 
    
let empty = ""

let exponent_marker = 'e' | 's' | 'f' | 'd' | 'l'
let sign = empty | '+'|'-'
let exactness = empty | "#i" | "#e"
let radix2 = "#b"
let radix8 = "#o"
let radix10 = empty | "#d"
let radix16 = "#x"
let digit2 = ['0'-'1']
let digit8 = ['0'-'7']
let digit10 = digit
let digit16 = ['0'-'9' 'a'-'f']
let suffix = empty | exponent_marker sign digit10+
let prefix2 = radix2 exactness | exactness radix2
let prefix8 = radix8 exactness | exactness radix8
let prefix10 = radix10 exactness | exactness radix10
let prefix16 = radix16 exactness | exactness radix16
let uinteger2 = digit2+ '#'*
let uinteger8 = digit8+ '#'*
let uinteger10= digit10+ '#'*
let integer10= sign digit10+ '#'*
let uinteger16 = digit16+ '#'*
let decimal10 = uinteger10 suffix | '.' digit10+ '#'* suffix 
   | digit10+ '.' digit10* '#'* suffix
   | digit10+ '#'+ '.' '#'* suffix
let ureal2 = uinteger2 | uinteger2 '/' uinteger2 '.' digit2+  
let ureal8 = uinteger8 | uinteger8 '/' uinteger8 '.' digit8+  
let ureal10 = uinteger10 | uinteger10 '/' uinteger10 '.' digit10+  | decimal10 
let ureal16 = uinteger16 | uinteger16 '/' uinteger16 '.' digit16+  
let real2 = sign ureal2
let real8 = sign ureal8
let real10 = sign ureal10
let real16 = sign ureal16
let complex2 = real2 | real2 '@' real2 | real2 '+' ureal2 'i' 
   | real2 '-' ureal2 'i' | real2 '+' 'i' | real2 '-' 'i'
   | '+' ureal2 'i' | '-' ureal2 'i' | '+' 'i' | '-' 'i' 
let complex8 = real8 | real8 '@' real8 | real8 '+' ureal8 'i' 
   | real8 '-' ureal8 'i' | real8 '+' 'i' | real8 '-' 'i'
   | '+' ureal8 'i' | '-' ureal8 'i' | '+' 'i' | '-' 'i' 
let complex10 = real10 | real10 '@' real10 | real10 '+' ureal10 'i' 
   | real10 '-' ureal10 'i' | real10 '+' 'i' | real10 '-' 'i'
   | '+' ureal10 'i' | '-' ureal10 'i' | '+' 'i' | '-' 'i' 
let complex16 = real16 | real16 '@' real16 | real16 '+' ureal16 'i' 
   | real16 '-' ureal16 'i' | real16 '+' 'i' | real16 '-' 'i'
   | '+' ureal16 'i' | '-' ureal16 'i' | '+' 'i' | '-' 'i' 
let num2 = prefix2 complex2
let num8 = prefix8 complex8
let num10 = prefix10 complex10
let num16 = prefix16 complex16
let number = num2 | num8 | num10 | num16


(*let number = [ '+' '-']?digit+  ȤꤢϤ *)
 
rule token = parse
| intertoken_space_nl { line := !line + 1; token lexbuf }
| intertoken_space{ token lexbuf }
| boolean       { BOOLEAN   ( (Lexing.lexeme lexbuf) = "#t" ) }
| number        { NUMBER    ( lex_num (Lexing.from_string (Lexing.lexeme lexbuf))) }
| character     { let str = Lexing.lexeme lexbuf in
                  match (String.length str) with
                    0|1 -> failwith "lexer error : character"
                  | 2   -> CHARACTER ' '
		  | 3   -> CHARACTER str.[2]
                  | _   -> 
                      begin
			match str with
			  "#\\space" -> CHARACTER ' '
                        | "#\\newline" -> CHARACTER '\n'
                        | _ -> failwith "lexer error : character"
                      end 
                }
| string        { STRING ( Lexing.lexeme lexbuf) }
| '('           { LPAREN }
| ')'           { RPAREN }
| "#("          { SHARP_LPAREN }
| "'"           { S_QUOTE }
| "`"           { S_UNQUOTE }
| ","           { S_COMMA }
| ",@"          { S_COMMA_ATMARK }
| "."           { DOT }
| "else"        { ELSE }
| "=>"          { RIGHTARROW }
| "define"      { DEFINE }
| "unquote"     { UNQUOTE }
| "unquote-splicing" { UNQUOTE_SPLICING } 
| "quote"       { QUOTE }
| "lambda"      { LAMBDA }
| "if"          { IF }
| "set!"        { SET }
| "begin"       { BEGIN }
| "cond"        { COND }
| "and"         { AND }
| "or"          { OR }
| "case"        { CASE }
| "let"         { LET }
| "let*"        { LETSTAR }
| "letrec"      { LETREC }
| "let-syntax"  { LET_SYNTAX }
| "letrec-syntax" { LETREC_SYNTAX }
| "define-syntax" { DEFINE_SYNTAX }
| "syntax-rules" { SYNTAX_RULES }
| "do"          { DO }
| "delay"       { DELAY }
| "quasiquote"  { QUASIQUOTE }
| identifier    { VARIABLE(Lexing.lexeme lexbuf) }
| eof           { EOF }
| _             { failwith ( "Lexer , line " ^ (string_of_int !line) ^ 
			     " , unknown token : " ^ Lexing.lexeme lexbuf) } 
and lex_num = parse
| integer10 { S_Int(int_of_string (Lexing.lexeme lexbuf))}
| real10 { S_Real(float_of_string (Lexing.lexeme lexbuf))}
| _      { failwith ( "Lexer , line " ^ (string_of_int !line) ^ 
		      " , unknown token : " ^ Lexing.lexeme lexbuf) } 
