(* Soft Type Ϣ *)

(* ----------------------------------- *)
(*  *)
(* ----------------------------------- *)

open Mylist


(* ٥ *)
type id = string

(*  *)
type stype = 
    { label:id; 
      num:bool; 
      bool:bool; 
      char:bool;
      port:bool;
      string:bool;
      symbol:bool;
      null:bool;
      mutable func:((stype list * stype) * stype) option;
      mutable pair:(stype * stype) option;
      mutable vector:stype option;
    }

(*  *)
type subst = (id * stype) list

(*  *)
type stype_scheme = Scheme of id list * stype
(* Ķ *)
type type_env = (string * stype_scheme) list




(* ----------------------------------- *)
(* ؿ *)
(* ----------------------------------- *)

let stype_car_of_pair st = 
  match st.pair with Some(car,_) -> car | _ -> failwith "stype:stype_car_of_pair"

(* ---------------------------------------------- *)
(* ʸɽؿ *)
(* ---------------------------------------------- *)

(* stype ʸɽ *)
let string_of_stype (st:stype) : string = 
  let rec string_of_onelevel (donelst:id list) (st:stype) : string =
    if(List.mem st.label donelst)
    then 
      st.label
    else
      begin 
	("(" ^ st.label^".") ^ 
	(if(st.num) then "int" else "") ^
	(if(st.bool) then "bool" else "") ^
	(if(st.char) then "char" else "") ^
	(if(st.string) then "string" else "") ^
	(if(st.symbol) then "symbol" else "") ^
	begin 
	  match st.func with
	    Some((args,tail),r) -> 
	      (List.fold_right (^)
		 (List.map (string_of_onelevel (st.label::donelst)) args)
		 "")
	      ^ ". " ^ (string_of_onelevel (st.label::donelst) tail)
	      ^ "->" ^ (string_of_onelevel (st.label::donelst) r)
	  | None -> ""
	end ^ 
	begin 
	  match st.pair with
	    Some(l,r) -> 
	      (string_of_onelevel (st.label::donelst) l)
	      ^ "*" ^
	      (string_of_onelevel (st.label::donelst) r)
	  | None -> ""
	end ^
	begin 
	  match st.vector with
	    Some(v) -> 
	      "#" ^ (string_of_onelevel (st.label::donelst) v)
	  | None -> ""
	end ^
	")"
      end
  in string_of_onelevel [] st

(* stype_scheme ʸɽ *)
let string_of_stype_scheme (sc:stype_scheme) : string = 
  match sc with
    Scheme(fall, st) -> "\\" ^ (List.fold_right (^) fall "") ^ "." ^ (string_of_stype st)

(* ɽ *)
let print_subst ((s:id),(st:stype)) :unit = 
  print_string (s ^ " -> " ^ st.label ^"\n")
let print_substs (s:subst) = List.iter print_subst s

(*----------------------------------------------------------*)
      

(* ٥ *)
let newlabel = 
  let n = ref 0 in 
  function () -> n := !n + 1 ; "t" ^ (string_of_int !n)

(*  *)
let newstype () = { label = newlabel (); 
		    num =false; 
		    bool=false;
		    char=false;
		    port=false;
		    string=false;
		    symbol=false;
		    null=false;
		    func = None;
		    pair = None;
		    vector = None;
		  }
(* 줾ο *)
let unknown_type () = { label = newlabel (); num =false; bool=false; port=false;
			char=false; string=false; symbol=false; null=false;
			func = None;  pair = None;  vector = None;}
let num_type () = { label = newlabel (); num =true; bool=false; port=false;
		    char=false; string=false; symbol=false; null=false;
		    func = None;  pair = None;  vector = None;}
let bool_type () = { label = newlabel (); num =false; bool=true; port=false;
		    char=false; string=false; symbol=false; null=false;
		     func = None;  pair = None;  vector = None;}
let null_type () = { label = newlabel (); num =false; bool=false; port=false;
		    char=false; string=false; symbol=false; null=true;
		     func = None;  pair = None;  vector = None;}
let port_type () = { label = newlabel (); num =false; bool=false; port=true;
		    char=false; string=false; symbol=false; null=false;
		     func = None;  pair = None;  vector = None;}
let char_type () = { label = newlabel (); num =false; bool=false; port=false;
		    char=true; string=false; symbol=false; null=false;
		     func = None;  pair = None;  vector = None;}
let string_type () = { label = newlabel (); num =false; bool=false; port=false;
		    char=false; string=true; symbol=false; null=false;
		     func = None;  pair = None;  vector = None;}
let symbol_type () = { label = newlabel (); num =false; bool=false; port=false;
		    char=false; string=false; symbol=true; null=false;
		     func = None;  pair = None;  vector = None;}
let func_type a b = { label = newlabel (); num =false; bool=false; port=false;
		    char=false; string=false; symbol=false; null=false;
		      func = Some(a,b);  pair = None;  vector = None;}
let pair_type a b = { label = newlabel (); num =false; bool=false; port=false;
		    char=false; string=false; symbol=false; null=false;
		      func = None;  pair = Some(a,b);  vector = None;}
let list_type a = 
  let tmpT = { label = newlabel (); num =false; bool=false; port=false;
	       char=false; string=false; symbol=false; null=true;
	       func = None;  pair = None;  vector = None;}
  in tmpT.pair <- Some(a,tmpT); tmpT


(* ˸ѿΥꥹȤõ *)
let fv_stype (st:stype) :id list =
  let rec fv_onelevel (donelst:id list) (st:stype) :id list =
    if(List.mem st.label donelst) then []
    else 
      st.label ::
      (uniq
	 (
	  begin
	    match st.func with
	      Some((arg,tail),r) -> 
		(List.flatten (List.map (fv_onelevel (st.label::donelst)) arg))
		@(fv_onelevel (st.label::donelst) tail)
		@(fv_onelevel (st.label::donelst) r)
	    | None -> []
	  end 
	  @
	    begin
	      match st.pair with
		Some(l,r) -> 
		  (fv_onelevel  (st.label::donelst) l)
		  @(fv_onelevel  (st.label::donelst) r)
	      | None -> []
	    end)
      )
  in fv_onelevel [] st


(* substitute  *)
let substitute (s:subst) (t:stype) =
  let rec subst_onelevel (donelst:(id * stype) list) (s:subst) (t:stype)  = 
    if(List.mem_assoc t.label donelst) 
    then List.assoc t.label donelst
    else
      if(List.mem_assoc t.label s)
      then 
	let subT = List.assoc t.label s in
	(* print_endline ("ppp" ^ " " ^ t.label ^ " -> " ^ subT.label); *)
	let newT = 
	  { label = subT.label;
	    num = t.num || subT.num;
	    bool = t.bool || subT.bool;
	    char= t.char || subT.char; 
	    port= t.port || subT.port; 
	    string=t.string || subT.string; 
	    symbol=t.symbol || subT.symbol; 
	    null=t.null || subT.null; 
	    func = None;
	    pair = None;
	    vector = None;
	  } 
	in 
	(* ƵŪʹ¤ϡȤη˻ĤäƤͥ褹롣*)
	(* polymorphsm Τ *)
	let func = 
	  begin 
	    match (t.func, subT.func) with
	      (Some((args1,tail1),ret1), Some((args2,tail2),ret2)) ->
(*		Printf.printf "ppp %s(%d) -> %s(%d)\n" t.label (List.length args1) subT.label (List.length args2);*)
		let newArgs = 
		  (List.map 
		     (subst_onelevel ((t.label,newT)::donelst) s) 
		     (Mylist.sub args1 0 (List.length args2)))
		in 
		Some((newArgs,
		      subst_onelevel ((t.label,newT)::donelst) s tail1),
		     subst_onelevel ((t.label,newT)::donelst) s ret1)	
	    | (None,subTfunc) -> subTfunc
	    | (Some((args1,tail1),ret1), None) -> 
		Some((List.map (subst_onelevel ((t.label,newT)::donelst) s) args1,
		      subst_onelevel ((t.label,newT)::donelst) s tail1),
		     subst_onelevel ((t.label,newT)::donelst) s ret1 )	
	  end 
	in
	newT.func <- func;
	let pair = 
	  begin
	    match (t.pair,subT.pair) with
	      (Some(a,b), _)-> Some(subst_onelevel ((t.label,newT)::donelst) s a,
			       subst_onelevel ((t.label,newT)::donelst) s b)
	    | (None,subTpair) -> subTpair
	  end 
	in 
	newT.pair <- pair;
	let vector = 
	  begin
	    match (t.vector,subT.vector) with
	      (Some(v), _)-> Some(subst_onelevel ((t.label,newT)::donelst) s v)
	    | (None,subTvec) -> subTvec
	  end 
	in 
	newT.vector <- vector;
(*	Printf.printf "ppp %s -> %s\n" t.label subT.label;
	print_endline (List.fold_right (^) (fv_stype newT) ""); *)
	newT
      else 
	let newT = 
	  { label = t.label;
	    num = t.num;
	    bool = t.bool ;
	    char= t.char;
	    port= t.port;
	    string= t.string;
	    symbol = t.symbol;
	    null= t.null;
	    func = None;
	    pair = None;
	    vector = None;
	  } 
	  in 
	let newFunc = 
	  begin
	    match t.func with
	      Some((args,tail),b)-> 
		Some((List.map (subst_onelevel ((t.label,newT)::donelst) s) args ,
		      subst_onelevel ((t.label,newT)::donelst) s tail),	
		     subst_onelevel ((t.label,newT)::donelst) s b )	
	    | None -> None
	  end 
	in
	newT.func <- newFunc;
	let newPair = 
	    begin
	      match t.pair with
		Some(a,b)-> Some(subst_onelevel ((t.label,newT)::donelst) s a,
				 subst_onelevel ((t.label,newT)::donelst) s b)
	      | None -> None
	    end;
	in 
	newT.pair <- newPair;
	let newVec = 
	    begin
	      match t.vector with
		Some(v)-> Some(subst_onelevel ((t.label,newT)::donelst) s v)
	      | None -> None
	    end;
	in 
	newT.vector <- newVec;
	newT
  in 
  let result = subst_onelevel [] s t in
  result
    
(* ι  s1 Ƥ s2 ΤƱˤʤ *)
let subst_combine (s2:subst) (s1:subst) : subst =
  s2@(List.map 
	(function (label,t) -> (label, substitute s2 t)) s1)
	
let rec assert_subst subst : unit=
  match subst with
    (x,_)::tl ->
      if List.mem_assoc x tl
      then 
	begin print_substs subst;failwith "subst assert failure" end
      else assert_subst tl
  | [] -> ()
      

(* unification *)    
let unify (t1:stype) (t2:stype) :subst =
  (* print_endline "unify";
  print_endline (string_of_stype t1);
  print_endline (string_of_stype t2);
  print_endline ""; *)
  let rec in_unify (t1:stype) (t2:stype) (donelst:(id * id) list) :subst =
    if((List.mem (t1.label,t2.label) donelst) 
     || (List.mem (t2.label,t1.label) donelst)
     || (t1.label = t2.label)) 
    then []
    else 
      let newstype = 
	{ label = newlabel ();
	  num =  t1.num  || t2.num;
	  bool = t1.bool || t2.bool;
	  char= t1.char || t2.char; 
	  port= t1.port || t2.port; 
	  string=t1.string || t2.string; 
	  symbol=t1.symbol || t2.symbol; 
	  null=t1.null || t2.null; 
	  func = None;
	  pair = None;
	  vector = None;
	}
      in 
      let newSubst = (t1.label, newstype)::(t2.label, newstype)::[] in
      let (funcS,funcT) = 
	(* ؿ unify *)
	match (t1.func, t2.func) with
	  (Some((args1,tail1),ret1), Some((args2,tail2),ret2)) ->
	    let rec unify_args subst args1 args2 = 
	      begin
		match (args1, args2) with
		  (hd1::tl1, hd2::tl2) ->
		    let tmpS = in_unify (substitute subst hd1) (substitute subst hd2)
			((t1.label,t2.label)::donelst) in
		    let (elseS, (elseT,tailT)) = 
		      unify_args (subst_combine tmpS subst) tl1 tl2 in
		    (elseS, ((substitute elseS hd1)::elseT,tailT))
		| ([],ts2) -> 
		    let tailS = unify_tails subst ts2  in
		    let tailS' = subst_combine tailS subst in
		    (tailS, ([],substitute tailS tail1))
		| (ts1,[])  ->
		    let tailS = subst_combine (unify_tails subst ts1) subst in
		    (tailS, ([],substitute tailS tail1))
	      end
	    and unify_tails subst remains = 
	      assert_subst subst;
	      begin
		match remains with
		  hd::tl ->
		    let tmpS = in_unify 
			(substitute subst hd) 
			(substitute subst (stype_car_of_pair tail1))
			((t1.label,t2.label)::donelst) in 
		    unify_tails (subst_combine tmpS subst) tl
		| [] -> 
		    let tmpS = in_unify 
			(substitute subst (stype_car_of_pair tail1))
			(substitute subst (stype_car_of_pair tail2))
			((t1.label,t2.label)::donelst) in 
		    let sss = subst_combine tmpS subst in
		    sss
	      end 
	    in
	    let (argS, (argT,tailT)) = unify_args newSubst args1 args2 in
	    let retS = in_unify (substitute argS ret1) (substitute argS ret2) 
		((t1.label,t2.label)::donelst) in
	    (* print_endline (if (newstype == (List.assoc t1.label retS)) then "ii" else "dame"); *)
	    let funcS = subst_combine retS argS in 
	    (funcS, Some((List.map (substitute funcS) argT,
			  substitute funcS tailT),
			 substitute funcS ret1))
	| (None, Some((arg2,tail2),ret2)) ->
	    (newSubst, Some((List.map (substitute newSubst) arg2,
			     substitute newSubst tail2),
			    substitute newSubst ret2))
	| (Some((arg1,tail2),ret1),None) ->
	    (newSubst, Some((List.map (substitute newSubst) arg1,
			     substitute newSubst ret1),
			    substitute newSubst ret1))
	| (None,None) ->
	    (newSubst,None)
      in 
      newstype.func <- funcT;
      let funcS = List.map 
	  (function (label,t) -> 
	    (label, substitute [(newstype.label,newstype)] t)) funcS
      in 
      (*(match (t1.func, t2.func) with
	 (Some _, Some _) -> 
	  print_endline (" -- " ^ (string_of_stype newstype));
	  print_endline (" --! " ^ (string_of_stype (List.assoc t1.label newSubst)));
	  print_endline (" --! " ^ (string_of_stype (List.assoc t1.label funcS)));
	  print_substs newSubst;
	  print_endline (if (newstype == (List.assoc t1.label funcS)) then "ii" else "dame"); 
	  print_substs funcS;
      | _ -> ());  *)

      (* pair unify *)
      let (pairS, pairT) = 
	match (t1.pair, t2.pair) with
	  (Some(car1, cdr1), Some(car2, cdr2)) ->
	    let carS = in_unify (substitute funcS car1) (substitute funcS car2)
		((t1.label,t2.label)::donelst) in
	    let tmpS = subst_combine carS funcS in 
	    let cdrS = in_unify (substitute tmpS cdr1) (substitute tmpS cdr2) 
		((t1.label,t2.label)::donelst) in
	    let pairS = subst_combine cdrS tmpS in 
	    (pairS ,Some(substitute pairS car1, substitute pairS cdr1))
	| (None, Some(car2,cdr2)) ->
	    (funcS, Some(substitute funcS car2, substitute funcS cdr2))
	| (Some(car1,cdr1),None) ->
	    (funcS, Some(substitute funcS car1, substitute funcS cdr1))
	| (None,None) ->
	    (funcS, None)
      in 
      newstype.pair <- pairT;
      let pairS = List.map 
	  (function (label,t) -> 
	    (label, substitute [(newstype.label,newstype)] t)) pairS
      in 
      (* vector unify *)
      let (vecS, vecT) = 
	match (t1.vector, t2.vector) with
	  (Some(v1), Some(v2)) ->
	    let vS = in_unify (substitute pairS v1) (substitute pairS v2)
		((t1.label,t2.label)::donelst) in
	    let vecS = subst_combine vS pairS in 
	    (vecS ,Some(substitute vecS v1))
	| (None, Some(v2)) ->
	    (pairS, Some(substitute pairS v2))
	| (Some(v1),None) ->
	    (pairS, Some(substitute funcS v1))
	| (None,None) ->
	    (pairS, None)
      in 
      newstype.vector <- vecT;
      let vecS = List.map 
	  (function (label,t) -> 
	    (label, substitute [(newstype.label,newstype)] t)) vecS
      in 
      vecS
  in 
  let s = in_unify t1 t2 [] in
  (*print_endline "result";
  print_endline (string_of_stype (List.assoc t1.label s));
  print_endline (string_of_stype (substitute s t1));
  print_endline (string_of_stype (substitute s t2));
  print_endline ""; *)
  s
    

(* ---------------------------------------------------------------- *)
(*  ͳѿ  generalize   *)
(* ---------------------------------------------------------------- *)

    
(* ͳѿ õ *)
let fv_scheme (sc:stype_scheme) : id list =
  match sc with
    Scheme(llst, st) -> 
      subtract_list (fv_stype st) llst
	
let fv_type_env (te:type_env) : id list = 
  let fvs = List.map (function (l,sc) -> fv_scheme sc) te in
  uniq (List.flatten fvs)
    
(* generalize *)
let generalize (te:type_env) (st:stype) : stype_scheme = 
  Scheme(subtract_list (fv_stype st) (fv_type_env te), st)


  
(* instantiate *)    
let instantiate (sc:stype_scheme) : stype = 
  match sc with
    Scheme(llst, st) ->
      let rec iter (map:(id * id) list) 
	  (donelst:(id*stype) list) (st:stype)  : stype = 
	if(List.mem_assoc st.label donelst) 
	then 
	  (* ̾դؤߤξ *)
	  List.assoc st.label donelst
	else 
	  (* forall «Ƥ뷿 *)
	  if(List.mem st.label llst) then
	    let newT = 
	      { label = List.assoc st.label map; 
		num = st.num; 
		bool = st.bool;
		char = st.char;
		port = st.port;
		string = st.string;
		symbol = st.symbol;
		null = st.null;
		func = None;
		pair = None;
		vector = None;
	      } 
	    in 
	    newT.func <- 
	      begin 
		match st.func with
		  Some((args,tail),ret) ->
		    Some((List.map (iter map ((st.label,newT)::donelst)) args,
			  iter map ((st.label,newT)::donelst) tail),
			 iter map ((st.label,newT)::donelst) ret)
		| None -> None
	      end;
	    newT.pair <-
	      begin 
		match st.pair with
		  Some(car,cdr) ->
		    Some(iter map ((st.label,newT)::donelst) car,
			 iter map ((st.label,newT)::donelst) cdr)
		| None -> None
	      end;
	    newT.vector <-
	      begin 
		match st.vector with
		  Some(v) ->
		    Some(iter map ((st.label,newT)::donelst) v)
		| None -> None
	      end;
	    newT
	  else 
	    (* forall «Ƥʤ *)
	    let newT = 
	      { label = st.label; 
		num =   st.num; 
		bool =  st.bool;
		char = st.char;
		port = st.port;
		string = st.string;
		symbol = st.symbol;
		null = st.null;
		func =  None; 
		vector =  None; 
		pair =  None 
	      }
	    in 
	    newT.func <-
	      begin
		match st.func with
		  Some((args,tail),ret) -> 
		    Some((List.map (iter map ((st.label,newT)::donelst)) args,
			  iter map ((st.label,newT)::donelst) tail),
			 iter map ((st.label,newT)::donelst) ret)
		| None -> None
	      end;
	    newT.pair <-
	      begin 
		match st.pair with
		  Some(car,cdr) ->
		    Some(iter map ((st.label,newT)::donelst) car,
			 iter map ((st.label,newT)::donelst) cdr)
		| None -> None
	      end;
	    newT.vector <-
	      begin 
		match st.vector with
		  Some(v) ->
		    Some(iter map ((st.label,newT)::donelst) v)
		| None -> None
	      end;
	    newT
	      
      in 
      let map = List.combine llst (List.map (function _ -> newlabel ()) llst) in
      iter map [] st
	
let stype2scheme st = Scheme([],st)       


(* instantiate *)    
let substitute_scheme (s:subst) (t:stype_scheme) =
  match t with
    Scheme(gen,st) ->
      if(List.fold_right (||)
	   (List.map (function x -> List.mem_assoc x s) gen) false)
      then 
	begin 
	  print_substs s; 
	  print_endline (string_of_stype_scheme t);
	  failwith "substitute_scheme error : " 
	end
      else Scheme(gen,substitute s st)

let substitute_tenv (s:subst) (tenv:type_env) : type_env =
  List.map (function (x,ts) -> (x,substitute_scheme s ts)) tenv
    
